"""
Highlighting for :class:`job widgets <upsies.uis.tui.widgets.JobWidget>`
"""

from prompt_toolkit import styles

# flake8: noqa: E241 multiple spaces after ','

# Remove defaults
styles.defaults.PROMPT_TOOLKIT_STYLE.clear()
styles.defaults.WIDGETS_STYLE.clear()

# https://python-prompt-toolkit.readthedocs.io/en/master/pages/advanced_topics/styling.html

style = styles.Style([
    ('default',                       ''),
    ('label',                         'bold'),

    ('output',                        ''),
    ('warning',                       'fg:#fe0 bold'),
    ('error',                         'fg:#f60 bold'),

    ('info',                          'bg:#222 fg:#dd5'),
    ('info.readout',                  'bg:#244 bold'),
    ('info.progressbar',              ''),
    ('info.progressbar.progress',     'reverse'),

    ('dialog',                        'bg:#222 fg:#5dd'),

    ('dialog.text',                   ''),

    ('dialog.choice',                 ''),
    ('dialog.choice.focused',         'reverse'),

    ('dialog.search',                 'bg:default'),
    ('dialog.search.label',           'bold underline'),
    ('dialog.search.query',           'bg:#222'),
    ('dialog.search.info',            'bg:#222'),
    ('dialog.search.results',         'bg:#222'),
    ('dialog.search.results.focused', 'reverse'),
])



"""
Create torrent file
"""

import collections
import datetime
import hashlib
import math
import os
import time
from os.path import exists as _path_exists

from .. import __project_name__, __version__, constants, errors, utils
from . import LazyModule, fs, types

torf = LazyModule(module='torf', namespace=globals())


def create(*, content_path, announce, source, torrent_path,
           init_callback, progress_callback,
           overwrite=False, exclude=()):
    """
    Generate and write torrent file

    :param str content_path: Path to the torrent's payload
    :param str announce: Announce URL
    :param str source: Value of the ``source`` field in the torrent. This makes
        the torrent unique for each tracker to avoid cross-seeding issues, so it
        is usually the tracker's abbreviated name.
    :param str torrent_path: Path of the generated torrent file
    :param str init_callback: Callable that is called once before torrent
        generation commences. It gets `content_path` as a tree where each node
        is a tuple in which the first item is the directory name and the second
        item is a sequence of `(file_name, file_size)` tuples.

        Example:

        .. code::

           ('Parent',
               ('Foo', (
                   ('Picture.jpg', 82489),
                   ('Music.mp3', 5315672),
                   ('More files', (
                       ('This.txt', 57734),
                       ('And that.txt', 184),
                       ('Also some of this.txt', 88433),
                   )),
               )),
               ('Bar', (
                   ('Yee.mp4', 288489392),
                   ('Yah.mkv', 3883247384),
               )),
           )
    :param str progress_callback: Callable that is called at regular intervals
        with a :class:`CreateTorrentStatus` object as a positional argument.
        Torrent creation is cancelled if `progress_callback` returns `True` or
        any other truthy value.
    :param bool overwrite: Whether to overwrite `torrent_path` if it exists
    :param exclude: Sequence of regular expressions that are matched against
        file system paths. Matching files are not included in the torrent.

    :raise TorrentError: if anything goes wrong

    :return: `torrent_path`
    """
    if not announce:
        raise errors.TorrentError('Announce URL is empty')
    if not source:
        raise errors.TorrentError('Source is empty')

    if overwrite or not _path_exists(torrent_path):
        torrent = None

        # Try to read piece hashes from cache
        torrent = _read_cache_torrent(
            content_path=content_path,
            exclude=exclude,
        )
        if torrent:
            # Update metadata in generic torrent
            torrent.trackers = (announce,)
            torrent.source = source
        else:
            # Reading cached piece hashes failed for some reason
            torrent = _generate_torrent(
                content_path=content_path,
                announce=announce,
                source=source,
                exclude=exclude,
                progress_callback=progress_callback,
                init_callback=init_callback,
            )
            if torrent.is_ready:
                _store_cache_torrent(torrent)

        # Write torrent file
        try:
            torrent.write(torrent_path, overwrite=True)
        except torf.TorfError as e:
            raise errors.TorrentError(e)

    return torrent_path


def _generate_torrent(*, content_path, announce, source, exclude, init_callback, progress_callback):
    callback = _CreateTorrentCallbackWrapper(progress_callback)
    try:
        torrent = torf.Torrent(
            path=content_path,
            exclude_regexs=exclude,
            trackers=((announce,),),
            private=True,
            source=source,
            created_by=f'{__project_name__} {__version__}',
            creation_date=time.time(),
        )
        init_callback(_make_file_tree(torrent.filetree))
        torrent.generate(callback=callback, interval=1.0)
    except torf.TorfError as e:
        raise errors.TorrentError(e)
    else:
        return torrent


def _store_cache_torrent(torrent):
    cache_torrent = torf.Torrent(
        private=True,
        created_by=f'{__project_name__} {__version__}',
        creation_date=time.time(),
        comment='This torrent is used to cache previously hashed pieces.',
    )
    _copy_torrent_info(torrent, cache_torrent)
    cache_torrent_path = _get_cache_torrent_path(cache_torrent, create_directory=True)
    cache_torrent.write(cache_torrent_path)


def _read_cache_torrent(content_path, exclude):
    try:
        torrent = torf.Torrent(
            path=content_path,
            exclude_regexs=exclude,
            private=True,
            created_by=f'{__project_name__} {__version__}',
            creation_date=time.time(),
        )
        cache_torrent_path = _get_cache_torrent_path(torrent, create_directory=False)
        cache_torrent = torf.Torrent.read(cache_torrent_path)
    except torf.TorfError:
        return None
    else:
        _copy_torrent_info(cache_torrent, torrent)
        return torrent


def _get_cache_torrent_path(torrent, create_directory=True):
    directory = os.path.join(
        constants.CACHE_DIRPATH,
        'generic_torrents',
    )
    if create_directory:
        try:
            fs.mkdir(directory)
        except errors.ContentError as e:
            raise errors.TorrentError(f'{directory}: {e}')

    cache_id = {
        'files': [(str(f), f.size) for f in torrent.files],
        'piece_size': torrent.piece_size,
    }
    cache_id_string = utils.semantic_hash(cache_id)
    filename = f'{torrent.name}.{cache_id_string}.torrent'
    return os.path.join(directory, filename)


def _copy_torrent_info(from_torrent, to_torrent):
    from_info, to_info = from_torrent.metainfo['info'], to_torrent.metainfo['info']
    to_info['pieces'] = from_info['pieces']
    to_info['piece length'] = from_info['piece length']
    to_info['name'] = from_info['name']
    if 'length' in from_info:
        to_info['length'] = from_info['length']
    else:
        to_info['files'] = from_info['files']


def _make_file_tree(tree):
    files = []
    for name,file in tree.items():
        if isinstance(file, collections.abc.Mapping):
            subtree = _make_file_tree(file)
            files.append((name, subtree))
        else:
            files.append((name, file.size))
    return tuple(files)


class _CreateTorrentCallbackWrapper:
    def __init__(self, callback):
        self._callback = callback
        self._progress_samples = []
        self._time_started = datetime.datetime.now()

    def __call__(self, torrent, filepath, pieces_done, pieces_total):
        time_now = datetime.datetime.now()
        percent_done = pieces_done / pieces_total * 100
        seconds_elapsed = time_now - self._time_started

        def progress_samples_trim_condition(samples):
            oldest_sample_age = time_now - samples[0][0]
            return oldest_sample_age > datetime.timedelta(seconds=10)

        self._maintain_samples(
            samples=self._progress_samples,
            new_sample=(time_now, pieces_done),
            trim_condition=progress_samples_trim_condition,
        )

        # Estimate how long torrent creation will take
        if len(self._progress_samples) >= 2:
            # Calculate the difference between each pair of samples
            diffs = [
                b[1] - a[1]
                for a, b in zip(self._progress_samples[:-1], self._progress_samples[1:])
            ]
            pieces_per_second = self._get_average(diffs, weight_factor=1.1)
            bytes_per_second = types.Bytes(pieces_per_second * torrent.piece_size)
            pieces_remaining = pieces_total - pieces_done
            bytes_remaining = pieces_remaining * torrent.piece_size
            seconds_remaining = datetime.timedelta(seconds=bytes_remaining / bytes_per_second)
        else:
            bytes_per_second = types.Bytes(0)
            seconds_remaining = datetime.timedelta(seconds=0)

        seconds_total = seconds_elapsed + seconds_remaining
        time_finished = self._time_started + seconds_total

        status = CreateTorrentStatus(
            bytes_per_second=bytes_per_second,
            filepath=filepath,
            percent_done=percent_done,
            piece_size=torrent.piece_size,
            pieces_done=pieces_done,
            pieces_total=pieces_total,
            seconds_elapsed=seconds_elapsed,
            seconds_remaining=seconds_remaining,
            seconds_total=seconds_total,
            time_finished=time_finished,
            time_started=self._time_started,
            total_size=torrent.size,
        )
        return self._callback(status)

    def _maintain_samples(self, new_sample, samples, trim_condition):
        samples.append(new_sample)
        while samples and trim_condition(samples):
            del samples[0]

    def _get_average(self, samples, weight_factor, get_value=lambda sample: sample):
        # Give recent samples more weight than older samples
        # https://en.wikipedia.org/wiki/Moving_average
        weights = []
        for _ in range(len(samples)):
            try:
                weight = weights[-1]
            except IndexError:
                weight = 1
            weights.append(weight * weight_factor)

        return sum(
            get_value(sample) * weight
            for sample, weight in zip(samples, weights)
        ) / sum(weights)


class CreateTorrentStatus(collections.namedtuple(
    typename='CreateTorrentStatus',
    field_names=(
        'bytes_per_second',
        'filepath',
        'percent_done',
        'piece_size',
        'pieces_done',
        'pieces_total',
        'seconds_elapsed',
        'seconds_remaining',
        'seconds_total',
        'time_started',
        'time_finished',
        'total_size',
    ),
)):
    """
    :func:`~.collections.namedtuple` with these attributes:

        - ``bytes_per_second`` (:class:`~.types.Bytes`)
        - ``filepath`` (:class:`str`)
        - ``percent_done`` (:class:`float`)
        - ``piece_size`` (:class:`~.types.Bytes`)
        - ``pieces_done`` (:class:`int`)
        - ``pieces_total`` (:class:`int`)
        - ``seconds_elapsed`` (:class:`float`)
        - ``seconds_remaining`` (:class:`float`)
        - ``seconds_total`` (:class:`float`)
        - ``time_started`` (:class:`float`)
        - ``time_finished`` (:class:`float`)
        - ``total_size`` (:class:`~.types.Bytes`)
    """


class TorrentFileStream:
    """
    Traverse files as they are described in a torrent

    :param torrent: :class:`~.torf.Torrent` object
    :param location: Path to location of files in `torrent`

    Instances should be used as a context manager to ensure all files are closed
    properly.

    Example:
    >>> torrent = torf.Torrent(...)
    >>> TorrentFileStream(torrent, location='path/to/files') as tfs:
    >>>     # Get the 29th piece of the concatenated file stream
    >>>     piece = tfs.get_piece(29)
    """

    def __init__(self, torrent, location):
        self._torrent = torrent
        self._location = location
        self._open_files = {}

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        for filepath, fh in tuple(self._open_files.items()):
            fh.close()
            del self._open_files[filepath]

    def get_file_position(self, file):
        """
        Return index of first byte of `file` in stream of concatenated files

        :param file: :class:`~torf.File` object
        """
        file_index = self._torrent.files.index(file)
        stream_pos = sum(f.size for f in self._torrent.files[:file_index])
        return stream_pos

    def get_file_at_position(self, byte_index):
        """
        Return file that belongs to the byte at `byte_index` in stream of
        concatenated files

        :param byte_index: Index of the byte; minimum is 0, maximum is the
            torrent's size minus 1
        """
        if byte_index >= 0:
            pos = 0
            for f in self._torrent.files:
                pos += f.size - 1
                if pos >= byte_index:
                    return f
                else:
                    pos += 1
        raise ValueError(f'byte_index is out of bounds (0 - {self._torrent.size}): {byte_index}')

    def get_piece_indexes_for_file(self, file):
        """
        Return indexes of pieces that contain at least one byte of `file`

        :param file: :class:`~torf.File` object
        """
        stream_pos = self.get_file_position(file)
        first_piece_index = math.floor(stream_pos / self._torrent.piece_size)
        last_piece_index = math.floor((stream_pos + file.size - 1) / self._torrent.piece_size)
        return list(range(first_piece_index, last_piece_index + 1))

    def get_files_for_byte_range(self, first_byte_index, last_byte_index):
        """
        Return list of files that have at least one byte at `first_byte_index`,
        `last_byte_index` or between those two in the stream of concatenated
        files
        """
        assert first_byte_index <= last_byte_index, (first_byte_index, last_byte_index)
        pos = 0
        files = []
        for file in self._torrent.files:
            file_first_byte_index = pos
            file_last_byte_index = pos + file.size - 1
            if (
                    # Is first byte of file inside of range?
                    first_byte_index <= file_first_byte_index <= last_byte_index or
                    # Is last byte of file inside of range?
                    first_byte_index <= file_last_byte_index <= last_byte_index or
                    # Are all bytes of file inside of range?
                    (first_byte_index >= file_first_byte_index and last_byte_index <= file_last_byte_index)
            ):
                files.append(file)
            pos += file.size
        return files

    def get_absolute_piece_indexes(self, file, relative_piece_indexes):
        """
        Return list of sanitized absolute piece indexes

        :param file: :class:`~torf.File` object
        :param relative_piece_indexes: Sequence of piece indexes within `file`;
            negative values address pieces at the end of `file`, e.g. [0, 12,
            -1, -2]

        Example:

        >>> # Assume `file` starts in the 50th piece in the stream of
        >>> # concatenated files and is 100 pieces long. `1000` and `-1000` are
        >>> # ignored because they are out of bounds.
        >>> tfs.validate_piece_indexes(file, (0, 1, 70, 75, 1000, -1000, -3, -2, -1))
        [50, 51, 120, 125, 147, 148, 149]
        """
        file_piece_indexes = self.get_piece_indexes_for_file(file)
        pi_abs_min = file_piece_indexes[0]
        pi_abs_max = file_piece_indexes[-1]
        pi_rel_min = 0
        pi_rel_max = pi_abs_max - pi_abs_min

        validated_piece_indexes = set()
        for pi_rel in relative_piece_indexes:
            pi_rel = int(pi_rel)

            # Convert negative to absolute index
            if pi_rel < 0:
                pi_rel = pi_rel_max - abs(pi_rel) + 1

            # Ensure relative piece_index is within bounds
            pi_rel = max(pi_rel_min, min(pi_rel_max, pi_rel))

            # Convert to absolute piece_index
            pi_abs = pi_abs_min + pi_rel
            validated_piece_indexes.add(pi_abs)

        return sorted(validated_piece_indexes)

    def get_piece(self, piece_index):
        """
        Return piece at `piece_index` or `None` if required files don't exist

        :param piece_index: Piece index within stream of concatenated files

        Errors from reading files existing (e.g. :class:`PermissionError`) are
        raised as :class:`~.errors.ContentError`.
        """
        piece_size = self._torrent.piece_size
        torrent_size = sum(f.size for f in self._torrent.files)

        min_piece_index = 0
        max_piece_index = math.floor((torrent_size - 1) / piece_size)
        if not min_piece_index <= piece_index <= max_piece_index:
            raise ValueError(f'piece_index must be in range {min_piece_index} - {max_piece_index}: {piece_index}')

        # Find out which files we need to read from
        first_byte_index_of_piece = piece_index * piece_size
        last_byte_index_of_piece = min(
            first_byte_index_of_piece + piece_size - 1,
            torrent_size - 1,
        )
        relevant_files = self.get_files_for_byte_range(first_byte_index_of_piece, last_byte_index_of_piece)

        # Find out where to start reading in the first relevant file
        if len(relevant_files) == 1:
            # Our piece belongs to a single file
            file_pos = self.get_file_position(relevant_files[0])
            seek_to = first_byte_index_of_piece - file_pos
        else:
            # Our piece is spread over multiple files
            file = self.get_file_at_position(first_byte_index_of_piece)
            file_pos = self.get_file_position(file)
            seek_to = file.size - ((file_pos + file.size) % piece_size)

        # Read piece data from `relevant_files`
        bytes_to_read = piece_size
        piece = bytearray()
        for f in relevant_files:
            fh = self._get_open_file(f)

            # We can't read the piece if any `relevant_file` is missing
            if fh is None:
                return None

            # It's theoretically possible that a differntly sized file can
            # produce the correct piece, but we assume the file is not what
            # we're looking for if it doesn't have the expected size.
            elif f.size != self._get_file_size_from_fs(f):
                return None

            try:
                fh.seek(seek_to)
                seek_to = 0

                content = fh.read(bytes_to_read)
                bytes_to_read -= len(content)
                piece.extend(content)
            except OSError as e:
                msg = e.strerror if e.strerror else e
                raise errors.ContentError(f'Failed to read from {f}: {msg}')

        # Ensure expected `piece` length
        if last_byte_index_of_piece == torrent_size - 1:
            exp_piece_size = torrent_size % piece_size
            if exp_piece_size == 0:
                exp_piece_size = piece_size
        else:
            exp_piece_size = piece_size
        assert len(piece) == exp_piece_size, (len(piece), exp_piece_size)
        return bytes(piece)

    def _get_file_size_from_fs(self, file):
        filepath = os.path.join(self._location, file)
        if os.path.exists(filepath):
            try:
                return os.path.getsize(filepath)
            except OSError:
                # Return None instead of raising an exception
                pass

    def _get_open_file(self, file):
        filepath = os.path.join(self._location, file)
        if filepath not in self._open_files:
            if not os.path.exists(filepath):
                return None
            else:
                try:
                    self._open_files[filepath] = open(filepath, 'rb')
                except OSError as e:
                    msg = e.strerror if e.strerror else e
                    raise errors.ContentError(f'Failed to open {filepath}: {msg}')
        return self._open_files.get(filepath, None)

    def get_piece_hash(self, piece_index):
        """
        Read piece at `piece_index` from file(s) and return its SHA1 hash

        :raise ContentError: if a file exists but cannot be read

        :return: :class:`bytes`
        """
        piece = self.get_piece(piece_index)
        if piece is not None:
            return hashlib.sha1(piece).digest()

    @property
    def torrent_piece_hashes(self):
        """Sequence of piece hashes stored in torrent"""
        pieces = self._torrent.metainfo['info']['pieces']
        piece_hashes = tuple(
            pieces[i : i + 20]
            for i in range(0, len(pieces), 20)
        )
        assert pieces[:20] == piece_hashes[0], (pieces[:20], piece_hashes[0])
        assert pieces[-20:] == piece_hashes[-1], (pieces[-20:], piece_hashes[-1])
        return piece_hashes

    def verify_piece(self, piece_index):
        """
        Generate SHA1 hash for piece at `piece_index` and compare to the expected
        hash in the torrent

        :raise ContentError: if a file exists but cannot be read

        :return: `True` if both hash values are equal, `False` otherwise
        """
        try:
            stored_piece_hash = self.torrent_piece_hashes[piece_index]
        except IndexError:
            max_piece_index = len(self.torrent_piece_hashes) - 1
            raise ValueError(f'piece_index must be in range 0 - {max_piece_index}: {piece_index}')

        generated_piece_hash = self.get_piece_hash(piece_index)
        if generated_piece_hash is not None:
            return stored_piece_hash == generated_piece_hash



"""
Find download path for a torrent, hardlinking existing files
"""

import collections
import functools
import os

from ..utils import LazyModule, fs, torrent
from . import JobBase

import logging  # isort:skip
_log = logging.getLogger(__name__)

torf = LazyModule(module='torf', namespace=globals())


class DownloadLocationJob(JobBase):
    """
    Find download path for a torrent, hardlinking existing files

    :param torrent: Path to torrent file
    :param locations: Sequence of directory paths that are searched for files in
        `torrent`
    :param default: Default download path if the files in `torrent` are not
        found
    """

    name = 'download-location'
    label = 'Download Location'
    hidden = True
    cache_id = None

    def initialize(self, *, torrent, locations, default=None):
        """
        Set internal state

        :param torrent: Path to torrent file
        """
        self._torrent_filepath = torrent
        # The torrent file is read in execute()
        self._torrent = None
        self._default_location = default
        self._locations = tuple(location for location in locations if location)
        # Map target paths to source paths
        self._links_created = {}

    def execute(self):
        if not self._locations:
            self.error('You must provide at least one location.')
        else:
            try:
                self._torrent = torf.Torrent.read(self._torrent_filepath)
            except torf.TorfError as e:
                self.error(e)
            else:
                target_location = self._get_target_location()
                if target_location is not None:
                    _log.debug('Outputting target location: %r', target_location)
                    self.send(target_location)
                elif self._default_location:
                    _log.debug('Outputting default location: %r', self._default_location)
                    self.send(self._default_location)
                else:
                    _log.debug('Outputting first location: %r', self._locations[0])
                    self.send(self._locations[0])
                self.finish()

    def _get_target_location(self):
        file_candidates = self._get_file_candidates()
        if file_candidates:
            _log.debug('### Finding piece-matching files')
        else:
            _log.debug('### Failed to find any size-matching files')

        links = {}
        target_location = None
        # `paths` maps relative file paths expected by torrent to a
        # candidate dictionary from `file_candidates`.
        for paths in self._each_set_of_linked_candidates(self._check_location, file_candidates):
            for file, candidate in paths.items():
                if file not in links:
                    if self._verify_file(file):
                        if target_location is None:
                            _log.debug('Setting target location: %r', candidate['location'])
                            target_location = candidate['location']

                        _log.debug('Using %r', candidate['filepath'])
                        links[file] = (candidate['filepath'], os.path.join(target_location, file))
                    else:
                        _log.debug('Not using %r', candidate['filepath'])
                else:
                    _log.debug('Already linked: %r', file)

        for source, target in links.values():
            self._create_hardlink(source, target)

        return target_location

    def _verify_file(self, file):
        _log.debug('Verifying content of %r at %r', file, self._check_location)
        with torrent.TorrentFileStream(self._torrent, self._check_location) as tfs:
            # Don't check the first and the last piece of a file as they likely
            # overlap with another file that might be either invalid or missing
            file_piece_indexes = tfs.get_absolute_piece_indexes(file, (1, -2))
            _log.debug('Verifying pieces: %r', file_piece_indexes)
            for piece_index in file_piece_indexes:
                piece_ok = tfs.verify_piece(piece_index)
                if piece_ok is True:
                    _log.debug('Piece %d is valid', piece_index)
                elif piece_ok is False:
                    _log.debug('Piece %d is invalid!', piece_index)
                    return False
                elif piece_ok is None:
                    _log.debug('Piece %d is unverifiable; assume non-existing file', piece_index)
                    return None
        return True

    def _each_set_of_linked_candidates(self, location, file_candidates):
        _log.debug('Iterating over sets of symlinks beneath %r', location)
        for pairs in _Combinator(file_candidates):
            links_created = []
            try:
                for file, candidate in pairs:
                    source = os.path.abspath(candidate['filepath'])
                    target = os.path.join(location, file)
                    self._create_symlink(source, target)
                    links_created.append((source, target))

                yield {file: candidate for file, candidate in pairs}

            finally:
                for source, target in links_created:
                    self._remove_link(target)

    def _get_file_candidates(self):
        # Map relative file paths expected by torrent to lists of files that
        # have the same size. A candidate is a dictionary that stores relevant
        # paths (see below).
        size_matching_files = collections.defaultdict(lambda: [])
        torrent_files = tuple(self._torrent.files)
        for filepath, location in self._each_file(*self._locations):
            for file in torrent_files:
                if self._is_size_match(file, filepath):
                    _log.debug('Size match for %r: %r', file, filepath)
                    #     file: Relative path within torrent
                    # location: Download path to give to the BitTorrent client
                    #           along with the torrent file
                    # filepath: Existing file path with same size as `file`
                    size_matching_files[file].append({
                        'location': location,
                        'filepath': filepath,
                    })
        return size_matching_files

    def _each_file(self, *paths):
        # Yield (filepath, path) tuples where the first item is a file (more
        # specifically: a non-directory) beneath the second item. The second
        # item is a path from `paths`. If there is a file path in `paths`, it is
        # yielded as both items of the tuple.
        for path in paths:
            _log.debug('Searching %s', path)
            if not os.path.isdir(path):
                yield str(path), str(path)
            else:
                for root, dirnames, filenames in os.walk(path, followlinks=True):
                    for filename in filenames:
                        yield str(os.path.join(root, filename)), str(path)

    @functools.lru_cache(maxsize=None)
    def _is_size_match(self, torrentfile, filepath):
        # Return whether a file in a torrent and an existing file are the same size
        return torrentfile.size == fs.file_size(filepath)

    def _create_hardlink(self, source, target):
        return self._create_link(os.link, source, target)

    def _create_symlink(self, source, target):
        return self._create_link(os.symlink, source, target)

    def _create_link(self, create_link_function, source, target):
        if not os.path.exists(target):
            # Create parent directory if it doesn't exist
            target_parent = os.path.dirname(target)
            try:
                # _log.debug(f'os.makedirs({target_parent!r}, exist_ok=True)')
                os.makedirs(target_parent, exist_ok=True)
            except OSError as e:
                msg = e.strerror if e.strerror else e
                self.error(f'Failed to create directory: {target_parent}: {msg}')
            else:
                # Create hardlink
                _log.debug(f'{create_link_function.__qualname__}({source!r}, {target!r})')
                try:
                    create_link_function(source, target)
                except OSError as e:
                    msg = e.strerror if e.strerror else e
                    self.error(f'Failed to create link: {target}: {msg}')
                else:
                    # Keep track of created links so that _remove_link() can
                    # refuse to delete anything we didn't create.
                    self._links_created[target] = source
        else:
            _log.debug('Already exists: %r', target)

    def _remove_link(self, target):
        # Only remove links we created ourselves.
        if target not in self._links_created:
            _log.debug('Refusing to delete %r: %r', target, self._links_created)
        elif os.path.exists(target):
            try:
                _log.debug(f'os.unlink({target!r})')
                os.unlink(target)
            except OSError as e:
                self.exception(e)
            else:
                del self._links_created[target]

                # Remove empty parent directories
                target_parent = os.path.dirname(target)
                try:
                    os.removedirs(target_parent)
                except OSError:
                    pass

    @property
    def _check_location(self):
        """Where to create temporary symlinks to verify pieces"""
        return os.path.join(self.cache_directory, self.name)


class _Combinator(collections.abc.Iterable):
    """
    Take a dictionary that maps keys to lists and turn it into a list of all
    possible ``(key, list item)`` combinations

    Example:

    >>> things = {
    >>>     'a': ['a1', 'a2'],
    >>>     'b': ['b1'],
    >>>     'c': ['c1', 'c2', 'c3', 'c4'],
    >>> }
    >>> _Combinator(things).combinations
    >>> [
    >>>     [('a', 'a1'), ('b', 'b1'), ('c', 'c1')],
    >>>     [('a', 'a1'), ('b', 'b1'), ('c', 'c2')],
    >>>     [('a', 'a1'), ('b', 'b1'), ('c', 'c3')],
    >>>     [('a', 'a1'), ('b', 'b1'), ('c', 'c4')],
    >>>     [('a', 'a2'), ('b', 'b1'), ('c', 'c1')],
    >>>     [('a', 'a2'), ('b', 'b1'), ('c', 'c2')],
    >>>     [('a', 'a2'), ('b', 'b1'), ('c', 'c3')],
    >>>     [('a', 'a2'), ('b', 'b1'), ('c', 'c4')],
    >>> ]
    """

    def __init__(self, candidates):
        self._candidates = {
            key: tuple(cands)
            for key, cands in candidates.items()
        }
        self._keys = tuple(self._candidates)
        self._indexes = {key: 0 for key in self._keys}

    def __len__(self):
        # Product of numbers of candidates
        length = 1
        for cands in self._candidates.values():
            length *= len(cands)
        return length

    def _advance(self):
        # Increase the rightmost index or reset it to zero
        # and increase the next index on the left
        for key in reversed(self._keys):
            if self._indexes[key] < len(self._candidates[key]) - 1:
                self._indexes[key] += 1
                break
            else:
                self._indexes[key] = 0

    @property
    def _pairs(self):
        # List of (key, list item) tuples for each key
        # using current list indexes
        pairs = []
        for key in self._keys:
            list_item = self._candidates[key][self._indexes[key]]
            pairs.append((key, list_item))
        return pairs

    def __iter__(self):
        combinations = []
        for _ in range(len(self)):
            combinations.append(self._pairs)
            self._advance()
        return iter(combinations)



"""
HTTP methods with caching
"""

import asyncio
import collections
import http
import io
import json
import os
import pathlib
import time

import httpx

from .. import __project_name__, __version__, constants, errors
from . import fs, html, semantic_hash

import logging  # isort:skip
_log = logging.getLogger(__name__)

_default_timeout = 60
_default_headers = {
    'Accept-Language': 'en-US,en;q=0.5',
    'User-Agent': f'{__project_name__}/{__version__}',
}

# Use one Lock per requested url + parameters so we can make multiple identical
# requests concurrently without bugging the server.
_request_locks = collections.defaultdict(lambda: asyncio.Lock())

# Use one single client session for all connections to benefit from re-using
# connections for multiple requests. The downside is that we have to remember to
# close the client before the application terminates.
_client = httpx.AsyncClient(
    timeout=_default_timeout,
    headers=_default_headers,
)


cache_directory = None
"""
Where to store cached requests

If this is set to a falsy value, default to :attr:`~.constants.CACHE_DIRPATH`.
"""


async def close():
    """
    Close the global client session used for all requests

    This coroutine function must be called before the application terminates.
    """
    _log.debug('Closing client: %r', _client)
    await _client.aclose()
    _log.debug('Closed client: %r', _client)


async def get(
        url,
        headers={},
        params={},
        auth=None,
        cache=False,
        max_cache_age=float('inf'),
        user_agent=False,
        follow_redirects=True,
        timeout=_default_timeout,
        cookies=None,
    ):
    """
    Perform HTTP GET request

    :param str url: URL to request
    :param dict headers: Custom headers (added to default headers)
    :param dict params: Query arguments, e.g. `{"k": "v"} → http://foo/bar?k=v`
    :param auth: Basic access authentication; sequence of <username> and
        <password> or `None`
    :param bool cache: Whether to use cached response if available
    :param int,float max_cache_age: Maximum age of cache in seconds
    :param bool user_agent: Whether to send the User-Agent header
    :param bool follow_redirects: Whether to follow redirects
    :param int,float timeout: Maximum number of seconds the request may take
    :param cookies: Cookies to include in the request (merged with
        existing cookies in the global client session)

        * If `cookies` is a :class:`dict`, forget them after the request.

        * If `cookies` is a :class:`str`, load cookies from that path (if it
          exists), perform the request, and save the new set of cookies to the
          same path.

        .. note:: Session cookies are not saved.

    :return: Response text
    :rtype: Response
    :raise RequestError: if the request fails for any expected reason
    """
    return await _request(
        method='GET',
        url=url,
        headers=headers,
        params=params,
        auth=auth,
        cache=cache,
        max_cache_age=max_cache_age,
        user_agent=user_agent,
        follow_redirects=follow_redirects,
        timeout=timeout,
        cookies=cookies,
    )

async def post(
        url,
        headers={},
        data={},
        files={},
        auth=None,
        cache=False,
        max_cache_age=float('inf'),
        user_agent=False,
        follow_redirects=True,
        timeout=_default_timeout,
        cookies=None,
    ):
    """
    Perform HTTP POST request

    :param str url: URL to request
    :param dict headers: Custom headers (added to default headers)
    :param dict data: Data to send as application/x-www-form-urlencoded
    :param dict files: Files to send as multipart/form-data as a dictionary that
        maps field names to file paths or dictionaries. For dictionaries, these
        keys are used:

             ``file``
                 File path or file-like object, e.g. return value of
                 :func:`open` or :class:`~.io.BytesIO`.

             ``filename`` (optional)
                 Name of the file that is reported to the server. Defaults to
                 :func:`~.os.path.basename` of ``file`` if ``file`` is a string,
                 `None` otherwise.

             ``mimetype`` (optional)
                 MIME type of the file content. If not provided and `file` is a
                 string, guess based on extension of ``file``. If `None`, do not
                 send any MIME type.

        Examples:

        >>> files = {
        >>>     "image": "path/to/foo.jpg",
        >>>     "document": {
        >>>         "file": "path/to/bar",
        >>>         "filename": "something.txt",
        >>>         "mimetype": "text/plain",
        >>>     },
        >>>     "data": {
        >>>         "file": io.BytesIO(b'binary data'),
        >>>         "mimetype": "application/octet-stream",
        >>>      },
        >>> })
    :param auth: Basic access authentication; sequence of <username> and
        <password> or `None`
    :param bool cache: Whether to use cached response if available
    :param int,float max_cache_age: Maximum age of cache in seconds
    :param bool user_agent: Whether to send the User-Agent header
    :param bool follow_redirects: Whether to follow redirects
    :param int,float timeout: Maximum number of seconds the request may take
    :param cookies: Cookies to include in the request (merged with existing
        cookies in the global client session); see :func:`get`

    :return: Response text
    :rtype: Response
    :raise RequestError: if the request fails for any expected reason
    """
    return await _request(
        method='POST',
        url=url,
        headers=headers,
        data=data,
        files=files,
        auth=auth,
        cache=cache,
        max_cache_age=max_cache_age,
        user_agent=user_agent,
        follow_redirects=follow_redirects,
        timeout=timeout,
        cookies=cookies,
    )

async def download(url, filepath, *args, **kwargs):
    """
    Write downloaded data to file

    :param url: Where to download the data from
    :param filepath: Where to save the downloaded data

    Any other arguments are passed to :func:`get`, except for `cache`, which is
    always `False`.

    If `filepath` exists, no request is made.

    :raise RequestError: if anything goes wrong
    :return: `filepath`
    """
    kwargs['cache'] = False
    if not os.path.exists(filepath):
        _log.debug('Downloading %r to %r', url, filepath)
        response = await get(url, *args, **kwargs)
        try:
            with open(filepath, 'wb') as f:
                f.write(response.bytes)
        except OSError as e:
            msg = e.strerror if e.strerror else str(e)
            raise errors.RequestError(f'Unable to write {filepath}: {msg}')
    else:
        _log.debug('Already downloaded %r to %r', url, filepath)
    return filepath


class Result(str):
    """
    Response to an HTTP request

    This is a subclass of :class:`str` with additional attributes and methods.
    """

    def __new__(cls, text, bytes, headers=None, status_code=None):
        obj = super().__new__(cls, text)
        obj._bytes = bytes
        obj._headers = headers if headers is not None else {}
        obj._status_code = status_code
        return obj

    @property
    def headers(self):
        """HTTP headers"""
        return self._headers

    @property
    def status_code(self):
        """HTTP status code"""
        return self._status_code

    @property
    def bytes(self):
        """Response data as :class:`bytes`"""
        return self._bytes

    def json(self):
        """
        Parse the response text as JSON

        :return: JSON as a `dict`
        :raise Requesterror: if parsing fails
        """
        try:
            return json.loads(self)
        except ValueError as e:
            raise errors.RequestError(f'Malformed JSON: {str(self)}: {e}')

    def __repr__(self):
        kwargs = [
            f'text={str(self)!r}',
            f'bytes={self.bytes!r}',
        ]
        if self.headers != {}:
            kwargs.append(f'headers={self.headers!r}')
        if self.status_code is not None:
            kwargs.append(f'status_code={self.status_code!r}')
        return f'{type(self).__name__}({", ".join(kwargs)})'


async def _request(
        method,
        url,
        headers={},
        params={},
        data={},
        files={},
        follow_redirects=True,
        cache=False,
        max_cache_age=float('inf'),
        auth=None,
        user_agent=False,
        timeout=_default_timeout,
        cookies=None,
    ):
    if method.upper() not in ('GET', 'POST'):
        raise ValueError(f'Invalid method: {method}')

    # Create request object
    if isinstance(data, (bytes, str)):
        build_request_args = {'content': data}
    else:
        build_request_args = {'data': data}
    request = _client.build_request(
        method=str(method),
        url=str(url),
        headers={**_default_headers, **headers},
        cookies=_load_cookies(cookies),
        params=params,
        files=_open_files(files),
        timeout=timeout,
        **build_request_args,
    )

    # Adjust User-Agent
    if isinstance(user_agent, str):
        request.headers['User-Agent'] = user_agent
    elif not user_agent:
        del request.headers['User-Agent']

    # Block when requesting the same URL simultaneously
    request_lock_key = (request.url, await request.aread())
    # _log.debug('Request lock key: %r', request_lock_key)
    request_lock = _request_locks[request_lock_key]
    async with request_lock:
        if cache:
            cache_file = _cache_file(method, url, params)
            result = _from_cache(cache_file, max_age=max_cache_age)
            if result is not None:
                return result

        _log.debug('%s: %r: %r', method, url, params)
        try:
            response = await _client.send(
                request=request,
                auth=auth,
                follow_redirects=follow_redirects,
            )
            try:
                response.raise_for_status()
            except httpx.HTTPStatusError:
                if response.status_code not in (301, 302, 303, 307, 308):
                    raise errors.RequestError(
                        f'{url}: {html.as_text(response.text)}',
                        url=url,
                        text=response.text,
                        headers=response.headers,
                        status_code=response.status_code,
                    )
        except httpx.TimeoutException:
            raise errors.RequestError(f'{url}: Timeout')
        except httpx.HTTPError as e:
            _log.debug(f'Unexpected HTTP error: {e!r}')
            raise errors.RequestError(f'{url}: {e}')
        else:
            if cookies and isinstance(cookies, (str, pathlib.Path)):
                _save_cookies(filepath=cookies, domain=response.url.host)

            if cache:
                cache_file = _cache_file(method, url, params)
                _to_cache(cache_file, response.content)

            return Result(
                text=response.text,
                bytes=response.content,
                headers=response.headers,
                status_code=response.status_code,
            )


def _load_cookies(cookies):
    if cookies:
        _log.debug('Loading cookies from %r', cookies)
    if isinstance(cookies, (collections.abc.Mapping, http.cookiejar.CookieJar)):
        return cookies
    elif cookies and isinstance(cookies, (str, pathlib.Path)):
        filepath_abs = os.path.join(constants.CACHE_DIRPATH, cookies)
        cookie_jar = http.cookiejar.LWPCookieJar(filepath_abs)
        try:
            cookie_jar.load()
        except FileNotFoundError:
            _log.debug('No such file: %r', filepath_abs)
            pass
        except OSError as e:
            msg = e.strerror if e.strerror else str(e)
            raise errors.RequestError(f'Failed to read {cookie_jar.filename}: {msg}')
        else:
            _log.debug('Loaded cookie jar: %r', cookie_jar)
        return cookie_jar
    elif cookies is not None:
        raise RuntimeError(f'Unsupported cookies type: {cookies!r}')


def _save_cookies(filepath, domain):
    filepath_abs = os.path.join(constants.CACHE_DIRPATH, filepath)
    _log.debug('Saving cookies for %r to %r', domain, filepath_abs)
    file_cookie_jar = http.cookiejar.LWPCookieJar(filepath_abs)
    for cookie in _client.cookies.jar:
        if cookie.domain.endswith(domain):
            _log.debug('Saving cookie: %r', cookie)
            file_cookie_jar.set_cookie(cookie)
        else:
            _log.debug('Ignoring cookie because %r does not end with %r: %r', cookie.domain, domain, cookie)

    if file_cookie_jar:
        try:
            file_cookie_jar.save()
        except OSError as e:
            msg = e.strerror if e.strerror else str(e)
            raise errors.RequestError(f'Failed to write {filepath_abs}: {msg}')
        else:
            _log.debug('Saved file cookie jar: %r', file_cookie_jar)


def _open_files(files):
    """
    Open files for upload

    See :func:`post` and
    https://www.python-httpx.org/advanced/#multipart-file-encoding.

    :raise RequestError: if a file cannot be opened
    """
    opened = {}
    for fieldname, fileinfo in files.items():
        if isinstance(fileinfo, str):
            filename = os.path.basename(fileinfo)
            fileobj = _get_file_object(fileinfo)
            opened[fieldname] = (filename, fileobj)

        elif isinstance(fileinfo, collections.abc.Mapping):
            if isinstance(fileinfo['file'], str):
                filename = fileinfo.get('filename', os.path.basename(fileinfo['file']))
                fileobj = _get_file_object(fileinfo['file'])
            elif isinstance(fileinfo['file'], io.IOBase):
                filename = fileinfo.get('filename', None)
                fileobj = fileinfo['file']
            else:
                raise RuntimeError(f'Invalid "file" value in fileinfo: {fileinfo["file"]!r}')
            mimetype = fileinfo.get('mimetype', False)
            if mimetype is False:
                opened[fieldname] = (filename, fileobj)
            else:
                opened[fieldname] = (filename, fileobj, mimetype)

        else:
            raise RuntimeError(f'Invalid fileinfo: {fileinfo}')

    return opened

def _get_file_object(filepath):
    """
    Open `filepath` and return the file object

    :raise RequestError: if OSError is raised by `open`
    """
    try:
        return open(filepath, 'rb')
    except OSError as e:
        msg = e.strerror if e.strerror else 'Failed to open'
        raise errors.RequestError(f'{filepath}: {msg}')


def _to_cache(cache_file, bytes):
    try:
        fs.mkdir(fs.dirname(cache_file))
        with open(cache_file, 'wb') as f:
            f.write(bytes)
    except OSError as e:
        raise RuntimeError(f'Unable to write cache file {cache_file}: {e}')


def _from_cache(cache_file, max_age=float('inf')):
    try:
        cache_mtime = os.stat(cache_file).st_mtime
    except FileNotFoundError:
        return None
    else:
        cache_age = time.time() - cache_mtime
        if cache_age <= max_age:
            bytes = _read_bytes_from_file(cache_file)
            text = _read_string_from_file(cache_file)
            if bytes and text:
                return Result(text=text, bytes=bytes)


def _read_string_from_file(filepath):
    bytes = _read_bytes_from_file(filepath)
    if bytes:
        return str(bytes, encoding='utf-8', errors='replace')


def _read_bytes_from_file(filepath):
    try:
        with open(filepath, 'rb') as f:
            return f.read()
    except OSError:
        pass


def _cache_file(method, url, params={}):
    def make_filename(method, url, params_str):
        if params_str:
            filename = f'{method.upper()}.{url}?{params_str}'
        else:
            filename = f'{method.upper()}.{url}'
        return filename.replace(' ', '+')

    if params:
        params_str = '&'.join((f'{k}={v}' for k,v in params.items()))
        if len(make_filename(method, url, params_str)) > 250:
            params_str = f'[HASH:{semantic_hash(params)}]'
    else:
        params_str = ''

    return os.path.join(
        cache_directory or constants.CACHE_DIRPATH,
        fs.sanitize_filename(make_filename(method, url, params_str)),
    )



import abc

from prompt_toolkit.filters import Condition
from prompt_toolkit.layout.containers import (ConditionalContainer, HSplit,
                                              Window, to_container)
from prompt_toolkit.layout.controls import FormattedTextControl
from prompt_toolkit.layout.layout import walk

from .. import widgets

import logging  # isort:skip
_log = logging.getLogger(__name__)


class JobWidgetBase(abc.ABC):
    """User-interaction and information display for :class:`~.jobs.JobBase` instance"""

    _no_runtime_widget = Window(
        dont_extend_height=True,
        style='class:info',
    )

    def __init__(self, job, app):
        self._job = job
        self._app = app
        self.setup()
        self.job.signal.register('info', lambda _: self.invalidate())
        self.job.signal.register('warning', lambda _: self.invalidate())
        main_widget = HSplit(
            children=[
                # Status information or user interaction
                ConditionalContainer(
                    filter=Condition(lambda: not self.job.is_finished),
                    content=self.runtime_widget or self._no_runtime_widget,
                ),
                # Result
                ConditionalContainer(
                    filter=Condition(lambda: self.job.output),
                    content=self.output_widget,
                ),
                # Additional info that isn't part of the job's main result
                # (e.g. CreateTorrentJobWidget can show the files in the
                # torrent, but the output is the torrent file path.)
                ConditionalContainer(
                    filter=Condition(lambda: not self.job.is_finished and bool(self.job.info)),
                    content=self.info_widget,
                ),
                # Warnings
                ConditionalContainer(
                    filter=Condition(lambda: bool(self.job.warnings)),
                    content=self.warnings_widget,
                ),
                # Errors
                ConditionalContainer(
                    filter=Condition(lambda: bool(self.job.errors)),
                    content=self.errors_widget,
                ),
            ],
        )
        label = widgets.HLabel(
            group='jobs',
            text=self.job.label,
            style='class:label',
            content=main_widget,
        )
        self._container = ConditionalContainer(
            filter=Condition(lambda: self.job.errors or not self.job.hidden),
            content=label,
        )

    @property
    def job(self):
        """Underlying :class:`~.JobBase` instance"""
        return self._job

    @abc.abstractmethod
    def setup(self):
        """
        Called on object creation

        Create widgets and register :attr:`job` callbacks.
        """

    @property
    @abc.abstractmethod
    def runtime_widget(self):
        """
        Interactive or status that is displayed while this job is running

        :return: :class:`~.prompt_toolkit.layout.containers.Window` object or
            `None`
        """

    @property
    def output_widget(self):
        """
        Job :attr:`~.JobBase.output`

        :return: :class:`~.prompt_toolkit.layout.containers.Window` object
        """
        return Window(
            style='class:output',
            # FIXME: If output is empty, prompt-toolkit ignores the
            #        "dont_extend_height" argument. Using a space (0x20) as a
            #        placeholder seems to prevent this issue.
            content=FormattedTextControl(lambda: '\n'.join(self.job.output) or ' '),
            dont_extend_height=True,
            wrap_lines=True,
        )

    @property
    def info_widget(self):
        """
        :attr:`~.JobBase.info` that is only displayed while this job is running

        :return: :class:`~.prompt_toolkit.layout.containers.Window` object
        """
        return Window(
            style='class:info',
            content=FormattedTextControl(lambda: str(self.job.info)),
            dont_extend_height=True,
            wrap_lines=True,
        )

    @property
    def warnings_widget(self):
        """
        Any :attr:`~.JobBase.warnings`

        :return: :class:`~.prompt_toolkit.layout.containers.Window` object
        """
        return Window(
            style='class:warning',
            content=FormattedTextControl(lambda: '\n'.join(str(e) for e in self.job.warnings)),
            dont_extend_height=True,
            wrap_lines=True,
        )

    @property
    def errors_widget(self):
        """
        Any :attr:`~.JobBase.errors`

        :return: :class:`~.prompt_toolkit.layout.containers.Window` object
        """
        return Window(
            style='class:error',
            content=FormattedTextControl(lambda: '\n'.join(str(e) for e in self.job.errors)),
            dont_extend_height=True,
            wrap_lines=True,
        )

    @property
    def is_interactive(self):
        """Whether this job needs user interaction"""
        if self.runtime_widget:
            for c in walk(to_container(self.runtime_widget), skip_hidden=True):
                if isinstance(c, Window) and c.content.is_focusable():
                    return True
        return False

    def invalidate(self):
        """Schedule redrawing of the TUI"""
        self._app.invalidate()

    def __pt_container__(self):
        return self._container



"""
Find most recent version
"""

import asyncio
import re

import natsort
from packaging.version import parse as parse_version

from .. import __project_name__, __version__
from . import http

import logging  # isort:skip
_log = logging.getLogger(__name__)


_PYPI_URL = f'https://pypi.org/pypi/{__project_name__}/json'
_REPO_URL = ('https://raw.githubusercontent.com/'
             f'plotski/{__project_name__}/master/{__project_name__}/__init__.py')
_MAX_CACHE_AGE = 12 * 3600  # 12 hours
_REQUEST_TIMEOUT = 3


async def get_newer_version():
    """
    Return the newest version if there is one

    If the current version is a prerelease, the return value may also be a
    prerelease.

    If the current version is a regular release, the returned version is also a
    regular release.

    :raise RequestError: if getting the latest version fails
    """
    current, release, prerelease = await _get_versions()
    current_parsed = parse_version(current)
    if current_parsed.is_prerelease:
        # Find newest release or prerelease, whichever is newer. Don't return
        # the parsed version because it removes padding zeros ("2021.06.20" ->
        # "2021.6.20").
        version_map = {
            parse_version(release): release,
            parse_version(prerelease): prerelease,
        }
        newest_parsed = sorted(version_map)[-1]
        if newest_parsed > current_parsed:
            return version_map[newest_parsed]
    else:
        if parse_version(release) > current_parsed:
            return release


async def _get_versions():
    current = __version__
    current_parsed = parse_version(current)
    if current_parsed.is_prerelease:
        gathered = await asyncio.gather(_get_latest_release(),
                                        _get_latest_prerelease())
        release, prerelease = gathered
        return current, release, prerelease
    else:
        release = await _get_latest_release()
        return current, release, None


async def _get_latest_release():
    response = await http.get(
        url=_PYPI_URL,
        timeout=_REQUEST_TIMEOUT,
        cache=True,
        max_cache_age=_MAX_CACHE_AGE,
    )
    all_versions = natsort.natsorted(response.json()['releases'])
    # PyPI should return the releases sorted by release date (latest last)
    if all_versions:
        return _fix_version(all_versions[-1])


async def _get_latest_prerelease():
    response = await http.get(
        url=_REPO_URL,
        timeout=_REQUEST_TIMEOUT,
        cache=True,
        max_cache_age=_MAX_CACHE_AGE,
    )
    match = re.search(r'^__version__\s*=\s*[\'"]([\w\.]+)[\'"]', response, flags=re.MULTILINE)
    if match:
        return match.group(1)


def _fix_version(version):
    match = re.search(r'^(\d{4})\.(\d+)\.(\d+)(.*)$', version)
    if match:
        year, month, day, pre = match.groups()
        version = f'{year}.{month:0>2}.{day:0>2}{pre}'
    return version



# """
# Image uploader for pixhost.to
# """

# from ...utils import html, http, types
# from .base import ImageHostBase

# import logging  # isort:skip
# _log = logging.getLogger(__name__)


# class PixhostImageHost(ImageHostBase):
#     """Upload images to pixhost.to"""

#     name = 'pixhost'

#     default_config = {
#         'base_url': 'https://api.pixhost.to',
#         'thumb_width': types.Integer(150, min=150, max=500),
#     }

#     # The file path is unique enough
#     cache_id = None

#     async def _upload(self, image_path):
#         return {
#             'url': 'foo',
#         }

#         # try:
#         #     response = await http.post(
#         #         url=self.options['base_url'] + '/images',
#         #         cache=False,
#         #         data={
#         #             'content_type': '1',
#         #             'max_th_size': self.options['thumb_width'],
#         #         },
#         #         files={
#         #             'img': image_path,
#         #         },
#         #     )
#         # except errors.RequestError as e:
#         #     # Reference: https://pixhost.to/api/index.html?shell#errors
#         #     if e.status_code == 413:
#         #         raise errors.RequestError('File size limit exceeded (10 MiB)')
#         #     else:
#         #         raise e

#         # else:
#         #     _log.debug('%s: Response: %r', self.name, response)
#         #     info = response.json()
#         #     _log.debug('%s: JSON: %r', self.name, info)
#         #     try:
#         #         return {
#         #             # We don't get an image URL; get it from the HTML
#         #             'url': self._get_image_url(info['show_url']),
#         #             'thumbnail_url': info['th_url'],
#         #         }
#         #     except KeyError:
#         #         raise RuntimeError(f'Unexpected response: {info}')

#     async def _get_image_url(self, html_url):
#         response = await http.get(html_url, cache=False)
#         doc = html.parse(response)
#         img_tag = doc.find('img', id='image')
#         _log.debug('Image tag: %r', img_tag)
#         if img_tag and img_tag.get('src'):
#             return img_tag['src']
#         else:
#             raise RuntimeError(f'Failed to find image: {html_url}')



"""
Scene release search and check
"""

from .. import errors
from ..utils import fs, scene, types
from . import JobBase

import logging  # isort:skip
_log = logging.getLogger(__name__)


class SceneSearchJob(JobBase):
    """
    Search for scene release

    This job adds the following signals to the :attr:`~.JobBase.signal`
    attribute:

        ``search_results``
            Emitted after new search results are available. Registered callbacks
            get a sequence of release names (:class:`str`) as a positional
            argument.
    """

    name = 'scene-search'
    label = 'Scene Search'
    hidden = True

    @property
    def cache_id(self):
        """Final segment of `content_path`"""
        return fs.basename(self._content_path)

    def initialize(self, *, content_path):
        """
        Set internal state

        :param content_path: Path to video file or directory that contains a
            video file or release name
        """
        self._content_path = content_path
        self.signal.add('search_results')

    def execute(self):
        """Send search query"""
        self.add_task(self._search(), finish_when_done=True)

    async def _search(self):
        try:
            query = scene.SceneQuery.from_string(self._content_path)
        except errors.SceneError as e:
            self.error(e)
        else:
            try:
                results = await scene.search(query)
            except (errors.RequestError, errors.SceneError) as e:
                self.error(e)
            else:
                _log.debug('Handling results: %r', results)
                if results:
                    for result in results:
                        self.send(result)
                else:
                    self.error('No results')
                self.signal.emit('search_results', results)
            finally:
                self.finish()


class SceneCheckJob(JobBase):
    """
    Verify scene release name and content

    This job adds the following signals to the :attr:`~.JobBase.signal`
    attribute:

        ``ask_release_name``
            Emitted if the user must pick a release name from multiple search
            results. Registered callbacks get a sequence of release names as
            positional argument.

        ``ask_is_scene_release``
            Emitted after we made our best guess and the user must approve or
            override it. Registered callbacks get a
            :class:`~.types.SceneCheckResult` enum as a positional argument.

        ``checked``
            Emitted after the user decided whether it's a scene release or not.
            Registered callbacks get a :class:`~.utils.types.SceneCheckResult`
            enum as a positional argument.
    """

    name = 'scene-check'
    label = 'Scene Check'
    hidden = False

    @property
    def cache_id(self):
        """Final segment of `content_path`"""
        return fs.basename(self._content_path)

    @property
    def is_scene_release(self):
        """
        :class:`~.utils.types.SceneCheckResult` enum or `None` before job is
        finished
        """
        return self._is_scene_release

    def initialize(self, *, content_path):
        """
        Set internal state

        :param content_path: Path to video file or directory that contains a
            video file or release name
        """
        self._content_path = content_path
        self._is_scene_release = None
        self._release_name = fs.strip_extension(fs.basename(content_path))
        self.signal.add('ask_release_name')
        self.signal.add('ask_is_scene_release')
        self.signal.add('checked')
        self.signal.record('checked')
        self.signal.register('checked', lambda is_scene: setattr(self, '_is_scene_release', is_scene))

    async def _catch_errors(self, coro):
        try:
            return await coro
        except errors.RequestError as e:
            self.warn(e)
            self.signal.emit('ask_is_scene_release', types.SceneCheckResult.unknown)
        except errors.SceneError as e:
            self.error(e)

    def execute(self):
        """Start asynchronous background worker task"""
        self.add_task(
            self._catch_errors(
                self._find_release_name()
            )
        )

    async def _find_release_name(self):
        is_scene_release = await scene.is_scene_release(self._content_path)
        if is_scene_release is types.SceneCheckResult.false:
            _log.debug('Not a scene release: %r', self._content_path)
            self._handle_scene_check_result(types.SceneCheckResult.false)
        else:
            # Find specific release name. If there are multiple search results,
            # ask the user to pick one.
            query = scene.SceneQuery.from_string(self._content_path)
            results = await scene.search(query, only_existing_releases=False)
            if not results:
                _log.debug('No search results: %r', self._content_path)
                self._handle_scene_check_result(types.SceneCheckResult.unknown)

            # Autopick if release name is in search results
            elif self._release_name in results:
                _log.debug('Autopicking from search results: %r', self._release_name)
                await self._verify_release(self._release_name)

            # Autopick single result if we know it's a scene release. If there
            # is missing information (e.g. no group), is_scene_release() returns
            # SceneCheckResult.unknown and we ask the user.
            elif len(results) == 1 and is_scene_release is types.SceneCheckResult.true:
                release_name = results[0]
                _log.debug('Autopicking only search result: %r', release_name)
                await self._verify_release(release_name)

            else:
                _log.debug('Asking for release name: %r', results)
                self.signal.emit('ask_release_name', results)

    def user_selected_release_name(self, release_name):
        """
        Must be called by the UI when the user picked a release name from search
        results

        :param release_name: Scene release name as :class:`str` or any falsy
            value for a non-scene release
        """
        _log.debug('User selected release name: %r', release_name)
        if release_name:
            self.add_task(
                self._catch_errors(
                    self._verify_release(release_name)
                )
            )
        else:
            self._handle_scene_check_result(types.SceneCheckResult.false)

    async def _verify_release(self, release_name):
        _log.debug('Verifying release: %r', release_name)
        is_scene_release, exceptions = await scene.verify_release(self._content_path, release_name)
        self._handle_scene_check_result(is_scene_release, exceptions)

    def _handle_scene_check_result(self, is_scene_release, exceptions=()):
        _log.debug('Handling result: %r: %r', is_scene_release, exceptions)

        warnings = [e for e in exceptions if isinstance(e, errors.SceneMissingInfoError)]
        for e in warnings:
            self.warn(e)

        serious_errors = [e for e in exceptions if not isinstance(e, errors.SceneMissingInfoError)]
        for e in serious_errors:
            self.error(e)

        if serious_errors:
            self.finish()
        elif is_scene_release in (types.SceneCheckResult.true, types.SceneCheckResult.false):
            self.finalize(is_scene_release)
        else:
            self.signal.emit('ask_is_scene_release', is_scene_release)

    def finalize(self, is_scene_release):
        """
        Make the final decision of whether this is a scene release or not and
        :meth:`finish`

        Must be called by the UI in the handler of the signal
        ``ask_is_scene_release``.

        :param is_scene_release: :class:`~.types.SceneCheckResult` enum
        """
        _log.debug('User decided: %r', is_scene_release)
        self.signal.emit('checked', is_scene_release)
        if is_scene_release is types.SceneCheckResult.true:
            self.send('Scene release')
        elif is_scene_release is types.SceneCheckResult.false:
            self.send('Not a scene release')
        else:
            self.send('May be a scene release')
        self.finish()



"""
Generate all required metadata and upload to tracker
"""

from .... import jobs, trackers, utils
from .base import CommandBase


class submit(CommandBase):
    """Generate all required metadata and upload to TRACKER"""

    names = ('submit',)

    argument_definitions = {}

    subcommand_name = 'TRACKER'
    subcommands = {
        tracker.name: {
            'description': f'Generate all required metadata and upload it to {tracker.label}.',
            'cli': {
                # Default arguments for all tackers
                **{
                    'CONTENT': {
                        'type': utils.argtypes.content,
                        'help': 'Path to release content',
                    },
                    ('--exclude-files', '--ef'): {
                        'nargs': '+',
                        'metavar': 'PATTERN',
                        'help': ('Glob pattern to exclude from torrent '
                                 '(matched case-sensitively against path in torrent)'),
                        'default': (),
                    },
                    ('--exclude-files-regex', '--efr'): {
                        'nargs': '+',
                        'metavar': 'PATTERN',
                        'help': ('Regular expression to exclude from torrent '
                                 '(matched case-sensitively against path in torrent)'),
                        'type': utils.argtypes.regex,
                        'default': (),
                    },
                    ('--add-to', '-a'): {
                        'type': utils.argtypes.client,
                        'metavar': 'CLIENT',
                        'help': ('Case-insensitive BitTorrent client name\n'
                                 'Supported clients: ' + ', '.join(utils.btclients.client_names())),
                    },
                    ('--copy-to', '-c'): {
                        'metavar': 'PATH',
                        'help': 'Copy the created torrent to PATH (file or directory)',
                    },
                },
                # Custom arguments defined by tracker for this command
                **tracker.TrackerConfig.argument_definitions.get('submit', {}),
            },
        }
        for tracker in trackers.trackers()
    }

    @utils.cached_property
    def jobs(self):
        submit_job = jobs.submit.SubmitJob(
            home_directory=self.home_directory,
            cache_directory=self.cache_directory,
            ignore_cache=self.args.ignore_cache,
            tracker=self.tracker,
            tracker_jobs=self.tracker_jobs,
        )
        return (
            tuple(self.tracker_jobs.jobs_before_upload)
            + (submit_job,)
            + tuple(self.tracker_jobs.jobs_after_upload)
        )

    @utils.cached_property
    def tracker_name(self):
        """Lower-case abbreviation of tracker name"""
        return self.args.subcommand.lower()

    @utils.cached_property
    def tracker_options(self):
        """
        :attr:`tracker_name` section in trackers configuration file combined with
        CLI arguments where CLI arguments take precedence unless their value is
        `None`
        """
        return self.get_options('trackers', self.tracker_name)

    @utils.cached_property
    def tracker(self):
        """
        :class:`~.trackers.base.TrackerBase` instance from one of the submodules of
        :mod:`.trackers`
        """
        return trackers.tracker(
            name=self.tracker_name,
            options=self.tracker_options,
        )

    @utils.cached_property
    def tracker_jobs(self):
        """
        :class:`~.trackers.base.TrackerJobsBase` instance from one of the submodules
        of :mod:`.trackers`
        """
        return self.tracker.TrackerJobs(
            options=self.tracker_options,
            content_path=self.args.CONTENT,
            tracker=self.tracker,
            image_host=self._get_imghost(),
            bittorrent_client=self._get_btclient(),
            torrent_destination=self._get_torrent_destination(),
            exclude_files=(
                tuple(self.args.exclude_files)
                + tuple(self.args.exclude_files_regex)
            ),
            common_job_args={
                'home_directory': self.home_directory,
                'cache_directory': self.cache_directory,
                'ignore_cache': self.args.ignore_cache,
            },
        )

    def _get_imghost(self):
        imghost_name = self.tracker_options.get('image_host', None)
        if imghost_name:
            # Get global image host options
            options = self.config['imghosts'][imghost_name].copy()

            # Apply tracker-specific options that are common for all image hosts
            options.update(self.tracker.TrackerJobs.image_host_config.get('common', {}))

            # Apply tracker-specific options for the used image host
            options.update(self.tracker.TrackerJobs.image_host_config.get(imghost_name, {}))

            return utils.imghosts.imghost(
                name=imghost_name,
                options=options,
            )

    def _get_btclient(self):
        btclient_name = (getattr(self.args, 'add_to', None)
                         or self.tracker_options.get('add-to', None)
                         or None)
        if btclient_name:
            return utils.btclients.client(
                name=btclient_name,
                config=self.config['clients'][btclient_name],
            )

    def _get_torrent_destination(self):
        return (getattr(self.args, 'copy_to', None)
                or self.tracker_options.get('copy-to', None)
                or None)



"""
Base class for image uploaders
"""

import abc
import collections
import copy
import os

from ... import __project_name__, constants, errors
from .. import fs, image
from . import common

import logging  # isort:skip
_log = logging.getLogger(__name__)


class ImageHostBase(abc.ABC):
    """
    Base class for image uploaders

    :param str cache_directory: Where to cache URLs; defaults to
        :attr:`.constants.CACHE_DIRPATH`
    :param options: User configuration options for this image host,
        e.g. authentication details, thumbnail size, etc
    :type options: :class:`dict`-like
    """

    def __init__(self, cache_directory=None, options=None):
        self._options = copy.deepcopy(self.default_config)
        self._options.update()
        if options is not None:
            self._options.update(options)
        self.cache_directory = cache_directory if cache_directory else constants.CACHE_DIRPATH

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        # This method is called for each subclass. This hack allows us to
        # overload `default_config` in subclasses without caring about common
        # defaults, e.g. subclasses don't need to have "thumb_width" in their
        # `default_config`, but it will exist anyway.
        cls.default_config = {
            **cls.default_config_common,
            **cls.default_config,
        }

    @property
    @abc.abstractmethod
    def name(self):
        """Name of the image hosting service"""

    @property
    def cache_directory(self):
        """Path to directory where upload info is cached"""
        return self._cache_dir

    @cache_directory.setter
    def cache_directory(self, directory):
        self._cache_dir = directory

    @property
    def options(self):
        """
        Configuration options provided by the user

        This is the :class:`dict`-like object from the initialization argument
        of the same name.
        """
        return self._options

    default_config_common = {
        'thumb_width': 0,  # No thumbnail
    }
    """Default user configuration for all subclasses"""

    default_config = {}
    """
    Default user configuration for a subclass

    This always contains :attr:`default_config_common`.
    """

    argument_definitions = {}
    """CLI argument definitions (see :attr:`.CommandBase.argument_definitions`)"""

    description = ''
    """Any documentation, for example how to get an API key"""

    async def upload(self, image_path, cache=True):
        """
        Upload image file

        :param image_path: Path to image file
        :param bool cache: Whether to attempt to get the image URL from cache

        :raise RequestError: if the upload fails

        :return: :class:`~.imghost.common.UploadedImage`
        """
        if 'apikey' in self.options and not self.options['apikey']:
            raise errors.RequestError(
                'You must configure an API key first. Run '
                f'"{__project_name__} upload-images {self.name} --help" '
                'for more information.'
            )

        info = {
            'url': await self._get_image_url(image_path, cache=cache),
        }

        thumb_width = self.options['thumb_width']
        if thumb_width:
            try:
                thumbnail_path = image.resize(
                    image_path,
                    width=thumb_width,
                    target_directory=self.cache_directory,
                )
            except errors.ImageResizeError as e:
                raise errors.RequestError(e)
            else:
                info['thumbnail_url'] = await self._get_image_url(thumbnail_path, cache=cache)

        return common.UploadedImage(**info)

    async def _get_image_url(self, image_path, cache=True):
        url = self._get_url_from_cache(image_path) if cache else None
        if not url:
            try:
                url = await self._upload_image(image_path)
            except errors.RequestError as e:
                raise errors.RequestError(e)
            else:
                _log.debug('Uploaded %r: %r', image_path, url)
                self._store_url_to_cache(image_path, url)
        else:
            _log.debug('Got URL from cache: %r: %r', image_path, url)
        return url

    @abc.abstractmethod
    async def _upload_image(self, image_path):
        """Upload `image_path` and return URL to the image file"""

    def _get_url_from_cache(self, image_path):
        cache_file = self._cache_file(image_path)
        if os.path.exists(cache_file):
            _log.debug('Already uploaded: %s', cache_file)
            try:
                with open(cache_file, 'r') as f:
                    return f.read().strip()
            except OSError:
                # We'll try to overwrite the bad cache file later
                pass

    def _store_url_to_cache(self, image_path, url):
        cache_file = self._cache_file(image_path)
        try:
            fs.mkdir(fs.dirname(cache_file))
            with open(cache_file, 'w') as f:
                f.write(url)
        except OSError as e:
            msg = e.strerror if getattr(e, 'strerror', None) else e
            raise RuntimeError(f'Unable to write cache {cache_file}: {msg}')

    def _cache_file(self, image_path):
        # If image_path is in our cache_directory, the image's file name makes
        # it unique. This is usually the case when we're uploading
        # screenshots. If image is not in our cache_directory, use the absolute
        # path as a unique identifier.
        if fs.dirname(image_path) == self.cache_directory:
            image_path = fs.basename(image_path)
        else:
            image_path = os.path.abspath(image_path)

        # Make cache file even more unique, e.g. imgbox only has one thumbnail
        # size per uploaded image, so if we want a different thumbnail size, we
        # need to upload again.
        cache_id = self._get_cache_id_as_string()
        if cache_id:
            image_path += f'.{cache_id}'

        # Max file name length is usually 255 bytes
        filename = fs.sanitize_filename(image_path[-200:]) + f'.{self.name}.url'

        return os.path.join(self.cache_directory, filename)

    def _get_cache_id_as_string(self):
        def as_str(obj):
            if isinstance(obj, collections.abc.Mapping):
                return ','.join(f'{as_str(k)}={as_str(v)}' for k,v in obj.items())
            elif isinstance(obj, str):
                return str(obj)
            elif isinstance(obj, collections.abc.Sequence):
                return ','.join(as_str(i) for i in obj)
            else:
                return str(obj)

        cache_id = self.cache_id
        if cache_id is not None:
            return as_str(cache_id)

    @property
    def cache_id(self):
        """
        Information that makes an upload unique, aside from the file path

        If this returns `None`, the file path is unique enough. Otherwise, the
        return value should be a string, dictionary, sequence or anything with a
        readable and unique string representation.
        """
        return None



"""
Abstract base class for commands
"""

import abc
import argparse
import collections
import os
import re
import sys
import textwrap

from .... import (__description__, __project_name__, __version__, constants,
                  defaults, errors, utils)
from .. import utils as tuiutils

import logging  # isort:skip
_log = logging.getLogger(__name__)


class _MyHelpFormatter(argparse.HelpFormatter):
    """Use "\n" to separate and preserve paragraphs and limit line width"""

    MAX_WIDTH = 90

    def __init__(self, *args, **kwargs):
        import shutil
        width_available = shutil.get_terminal_size().columns
        super().__init__(*args,
                         width=min(width_available, self.MAX_WIDTH),
                         **kwargs)

    def _fill_text(self, text, width, indent):
        return '\n'.join(self.__wrap_paragraphs(text, width))

    def _split_lines(self, text, width):
        return self.__wrap_paragraphs(text, width)

    def __wrap_paragraphs(self, text, width, indent=''):
        def wrap(text):
            if not text:
                return ['']
            else:
                return textwrap.wrap(
                    text=text,
                    width=min(width, self.MAX_WIDTH),
                    replace_whitespace=False,
                    initial_indent=indent,
                    subsequent_indent=indent,
                )

        if tuiutils.is_tty():
            text = re.sub(r'``(.*?)``', '\x1b[3m\\1\x1b[23m', text)
        else:
            text = re.sub(r'``(.*?)``', r'"\1"', text)

        return [line
                for paragraph in text.split('\n')
                for line in wrap(paragraph)]


class CommandBase(abc.ABC):
    """
    Base class for all commands

    :meth:`register` must be called on all subclasses.

    Instead of instantiating subclasses normally, the class method :meth:`run`
    should be used to get instances.

    :param args: Command line arguments
    :param config: :class:`~.configfiles.ConfigFiles` object
    """

    _argparser = argparse.ArgumentParser(
        description=__description__,
        formatter_class=_MyHelpFormatter,
    )
    _argparser.add_argument('--version',
                            action='version',
                            version=f'{__project_name__} {__version__}')
    _argparser.add_argument('--debug', '-d',
                            metavar='FILE',
                            help='Write debugging messages to FILE')
    _argparser.add_argument('--config-file', '-f',
                            help='General configuration file path',
                            default=constants.CONFIG_FILEPATH)
    _argparser.add_argument('--trackers-file', '-t',
                            help='Trackers configuration file path',
                            default=constants.TRACKERS_FILEPATH)
    _argparser.add_argument('--imghosts-file', '-i',
                            help='Image hosting services configuration file path',
                            default=constants.IMGHOSTS_FILEPATH)
    _argparser.add_argument('--clients-file', '-c',
                            help='BitTorrent clients configuration file path',
                            default=constants.CLIENTS_FILEPATH)
    _argparser.add_argument('--ignore-cache', '-C',
                            help='Ignore results from previous calls',
                            action='store_true')

    # Commands
    _subparsers = _argparser.add_subparsers(title='commands')

    # Mutually exclusive arguments
    _mutex_groups = collections.defaultdict(lambda: {})

    names = NotImplemented
    """
    Sequence of command names

    The first name is the full name and the rest are short aliases.
    """

    description = ''
    """
    Extended description

    The class docstring is the main description. Use this class attribute to
    generate text programmatically.
    """

    argument_definitions = NotImplemented
    """
    CLI argument definitions for this command

    This is a :class:`dict` in which keys are option names or flags
    (e.g. ``"PATH"`` or ``("--path", "-p")`` and each value is a :class:`dict`
    with keyword arguments for :meth:`argparse.ArgumentParser.add_argument`.

    Additionally, you may specify a ``group`` in each keyword argument
    dictionary. Arguments with the same ``group`` value are mutually exclusive,
    meaning the user can only specify one of them.
    """

    subcommands = {}
    """
    Subcommands of commands (or subsubcommands from a CLI point of view)

    This is a :class:`dict` where keys are subcommand names and values are
    :attr:`argument_definitions`.
    """

    subcommand_descriptions = {}
    """
    Descriptions of subcommands

    For each key in :attr:`subcommands`, this :class:`dict` may provide a longer
    multiline (or even multiparagraph) description.
    """

    subcommand_name = 'PLEASE_GIVE_THIS_SUBCOMMAND_A_PROPER_METAVAR'
    """Reference to expected subcommand in help texts"""

    @classmethod
    def register(cls):
        """
        Add the command and its arguments to the internal
        :class:`argparse.ArgumentParser` instance

        This classmethod must be called on every subclass.
        """
        # Get long text in output of "upsies <subcommand> --help"
        if cls.description:
            description = textwrap.dedent(cls.__doc__.strip('\n')) + '\n\n' + cls.description
        else:
            description = textwrap.dedent(cls.__doc__.strip('\n'))

        if cls.subcommands:
            description += f'\n\nRun ``upsies {cls.names[0]} {cls.subcommand_name} --help`` for more information.'

        # Get short text in output of "upsies --help"
        help = description.split('\n', 1)[0]

        parser = cls._subparsers.add_parser(
            cls.names[0],
            aliases=cls.names[1:],
            help=help,
            description=description,
            formatter_class=_MyHelpFormatter,
        )
        parser.set_defaults(command=cls)

        # Add options and flags for this command
        cls._add_args(parser, cls.argument_definitions, cls._mutex_groups[None])

        # Add subcommands
        if cls.subcommands:
            subparsers = parser.add_subparsers(
                metavar=cls.subcommand_name,
                help='Valid values: ' + ', '.join(sorted(cls.subcommands)),
            )
            # FIXME: When Python 3.6 is not supported anymore, `required=True`
            #        should be passed to add_subparsers() instead.
            #        https://docs.python.org/3/library/argparse.html#argparse.ArgumentParser.add_subparsers
            #        https://bugs.python.org/issue26510
            #        https://bugs.python.org/issue9253
            subparsers.required = True
            for subcmd_name, subcmd_info in cls.subcommands.items():
                description = subcmd_info.get('description', '')
                args = subcmd_info.get('cli', {})
                subparser = subparsers.add_parser(
                    subcmd_name,
                    description=description,
                    formatter_class=_MyHelpFormatter,
                )
                subparser.set_defaults(subcommand=subcmd_name)
                cls._add_args(subparser, args, cls._mutex_groups[subcmd_name])

    @staticmethod
    def _add_args(parser, argument_definitions, mutex_groups):
        for argname, argopts in argument_definitions.items():
            names = (argname,) if isinstance(argname, str) else argname
            group_name = argopts.pop('group', None)
            if group_name:
                # Put arguments with the same argopts["group"] in a mutually
                # exclusive group.
                # FIXME: https://bugs.python.org/issue41854
                if not names[0].startswith('-') and argopts.get('default') is None:
                    raise RuntimeError('Default values of mutually exclusive positional arguments '
                                       'must not be None. See: https://bugs.python.org/issue41854')
                if group_name not in mutex_groups:
                    mutex_groups[group_name] = parser.add_mutually_exclusive_group()
                mutex_groups[group_name].add_argument(*names, **argopts)

            else:
                # Allow using REMAINDER flag without importing argparse
                # everywhere. This also introduces an abstraction layer in case
                # we ever use something else than argparse.
                if argopts.get('nargs') == 'REMAINDER':
                    argopts['nargs'] = argparse.REMAINDER

                parser.add_argument(*names, **argopts)

    @classmethod
    def run(cls, args):
        """
        Execute command

        :param args: Sequence of CLI arguments or `None` to use `sys.argv`
        """
        try:
            if args is None:
                main_args, remaining_args = cls._argparser.parse_known_args()
            else:
                main_args, remaining_args = cls._argparser.parse_known_args(args)
        except SystemExit as e:
            # argparse has sys.exit(2) hardcoded for CLI errors.
            raise SystemExit(e.code if e.code in (0, 1) else 1)

        # Show help if no command given
        if not hasattr(main_args, 'command'):
            cls._argparser.print_help()
            raise SystemExit(0)

        # Debugging
        if main_args.debug:
            logging.basicConfig(
                format='%(asctime)s: %(name)s: %(message)s',
                filename=main_args.debug,
            )
            logging.getLogger(__project_name__).setLevel(level=logging.DEBUG)
            _log.debug('Started logging with %s version %s', __project_name__, __version__)

        # Read config files
        try:
            config = utils.configfiles.ConfigFiles(defaults=defaults.defaults)
            config.read('config', filepath=main_args.config_file, ignore_missing=True)
            config.read('trackers', filepath=main_args.trackers_file, ignore_missing=True)
            config.read('imghosts', filepath=main_args.imghosts_file, ignore_missing=True)
            config.read('clients', filepath=main_args.clients_file, ignore_missing=True)
        except errors.ConfigError as e:
            print(e, file=sys.stderr)
            raise SystemExit(1)

        return main_args.command(args=args, config=config)

    def __init__(self, args, config):
        self._args = self._argparser.parse_args(args)
        self._config = config

    @property
    @abc.abstractmethod
    def jobs(self):
        """
        Sequence of :class:`~.jobs.base.JobBase` objects

        For convenience, the sequence may also contain `None` instead.
        """

    @property
    def jobs_active(self):
        """Same as :attr:`jobs` but without `None` values"""
        return [j for j in self.jobs if j is not None]

    @property
    def args(self):
        """Parsed CLI arguments as :class:`argparse.Namespace` object"""
        return self._args

    @property
    def config(self):
        """Config file options as :class:`~.configfiles.ConfigFiles` object"""
        return self._config

    def get_options(self, config_name, config_section):
        """
        Combine section in config file with CLI arguments

        CLI arguments take precedence unless their value is `None`.

        :param str config_name: Config file name with out the "INI" extension
        :param str config_section: Name of a section in the config file

        :return: `dict`
        """
        config = self.config[config_name][config_section]
        args = vars(self.args)
        options = {}
        options.update(config)
        for k, v in args.items():
            if v is not None:
                options[k] = v
        return options

    @property
    def home_directory(self):
        """
        Passed as `home_directory` argument to :class:`.jobs.base.JobBase` instances
        that are instantiated by this class

        The default implementation passes the ``CONTENT`` or ``RELEASE``
        argument and ``config.main.cache_directory`` to :func:`.fs.projectdir`.

        If no ``CONTENT`` or ``RELEASE`` argument exists, return
        ``config.main.cache_directory`` directly.
        """
        if hasattr(self.args, 'CONTENT'):
            content_path = self.args.CONTENT
        elif hasattr(self.args, 'RELEASE'):
            content_path = self.args.RELEASE
        else:
            return self.config['config']['main']['cache_directory']

        return utils.fs.projectdir(
            content_path=content_path,
            base=self.config['config']['main']['cache_directory'],
        )

    @property
    def cache_directory(self):
        """
        Passed as `cache_directory` argument to :class:`.jobs.base.JobBase`
        instances that are instantiated by this class

        The default implementation appends ``.cache`` to :attr:`home_directory`
        if the ``CONTENT`` or ``RELEASE`` argument exists and returns
        ``config.main.cache_directory`` otherwise.
        """
        if hasattr(self.args, 'CONTENT') or hasattr(self.args, 'RELEASE'):
            return os.path.join(self.home_directory, '.cache')
        else:
            return self.config['config']['main']['cache_directory']



"""
Concrete :class:`~.base.TrackerConfigBase` subclass for bB
"""

import base64

from ...utils import argtypes, imghosts, types
from ..base import TrackerConfigBase


class BbTrackerConfig(TrackerConfigBase):
    defaults = {
        'base_url'    : base64.b64decode('aHR0cHM6Ly9iYWNvbmJpdHMub3Jn').decode('ascii'),
        'username'    : '',
        'password'    : '',
        'source'      : 'bB',
        'image_host'  : types.Choice('imgbox', options=(imghost.name for imghost in imghosts.imghosts())),
        'screenshots' : types.Integer(2, min=2, max=10),
        'exclude'     : [
            r'\.(?i:nfo|txt|jpg|jpeg|png|sfv|md5)$',
            r'/(?i:sample|extra|bonus|feature)',
            r'(?i:sample\.[a-z]+)$',
        ],
    }

    argument_definitions = {
        'submit': {
            ('--anime', '--an'): {
                'help': 'Tag upload as anime (ignored for movies)',
                'action': 'store_true',
            },
            ('--description', '--de'): {
                'help': 'Custom description as text or path to UTF-8 encoded text file',
            },
            ('--poster', '--po'): {
                'help': 'Path or URL to poster image',
            },
            ('--screenshots', '--ss'): {
                'help': ('How many screenshots to make '
                         f'(min={defaults["screenshots"].min}, '
                         f'max={defaults["screenshots"].max})'),
                'type': argtypes.number_of_screenshots(
                    min=defaults['screenshots'].min,
                    max=defaults['screenshots'].max,
                ),
            },
            ('--tags', '--ta'): {
                'help': 'Custom list of comma-separated tags',
            },
            ('--only-description', '--od'): {
                'help': 'Only generate description (do not upload anything)',
                'action': 'store_true',
                'group': 'generate-metadata',
            },
            ('--only-poster', '--op'): {
                'help': 'Only generate poster URL (do not upload anything)',
                'action': 'store_true',
                'group': 'generate-metadata',
            },
            ('--only-release-info', '--or'): {
                'help': 'Only generate release info (do not upload anything)',
                'action': 'store_true',
                'group': 'generate-metadata',
            },
            ('--only-tags', '--ota'): {
                'help': 'Only generate tags (do not upload anything)',
                'action': 'store_true',
                'group': 'generate-metadata',
            },
            ('--only-title', '--oti'): {
                'help': 'Only generate title (do not upload anything)',
                'action': 'store_true',
                'group': 'generate-metadata',
            },
        },
    }



"""
Abstract base class for scene release databases
"""

import abc
import copy


class SceneDbApiBase(abc.ABC):
    """Base class for scene release database APIs"""

    def __init__(self, config=None):
        self._config = copy.deepcopy(self.default_config)
        if config is not None:
            self._config.update(config.items())

    @property
    @abc.abstractmethod
    def name(self):
        """Unique name of the scene release database"""

    @property
    @abc.abstractmethod
    def label(self):
        """User-facing name of the scene release database"""

    @property
    def config(self):
        """
        User configuration

        This is a deep copy of :attr:`default_config` that is updated with the
        `config` argument from initialization.
        """
        return self._config

    @property
    @abc.abstractmethod
    def default_config(self):
        """Default user configuration as a dictionary"""

    async def search(self, query, only_existing_releases=True):
        """
        Search for scene release

        :param SceneQuery query: Search query
        :param bool only_existing_releases: See
            :meth:`~.scene.find.SceneQuery.search`

        :return: :class:`list` of release names as :class:`str`

        :raise RequestError: if the search request fails
        """
        return await query.search(
            self._search,
            only_existing_releases=only_existing_releases,
        )

    @abc.abstractmethod
    async def _search(self, keywords, group=None):
        pass

    @abc.abstractmethod
    async def release_files(self, release_name):
        """
        Map file names to file objects

        What a file object is depends on the subclass implementation.

        :param str release_name: Exact name of the release

        :raise RequestError: if request fails
        """



"""
Concrete :class:`~.base.TrackerConfigBase` subclass for BHD
"""

import base64

from ...utils import argtypes, imghosts, types
from ..base import TrackerConfigBase


class BhdTrackerConfig(TrackerConfigBase):
    defaults = {
        'upload_url'       : base64.b64decode('aHR0cHM6Ly9iZXlvbmQtaGQubWUvYXBpL3VwbG9hZA==').decode('ascii'),
        'announce_url'     : base64.b64decode('aHR0cHM6Ly90cmFja2VyLmJleW9uZC1oZC5tZToyMDUzL2Fubm91bmNl').decode('ascii'),
        'announce_passkey' : '',
        'apikey'           : '',
        'source'           : 'BHD',
        'anonymous'        : types.Bool('no'),
        'draft'            : types.Bool('no'),
        'image_host'       : types.Choice('imgbox', options=(imghost.name for imghost in imghosts.imghosts())),
        'screenshots'      : types.Integer(4, min=3, max=10),
        'exclude'          : [
            r'\.(?i:txt|jpg|jpeg|png|sfv|md5)$',
            r'/(?i:sample)',
        ],
    }

    argument_definitions = {
        'submit': {
            ('--custom-edition', '--ce'): {
                'help': 'Non-standard edition, e.g. "Final Cut"',
                'default': '',
            },
            ('--draft', '--dr'): {
                'help': 'Upload as draft',
                'action': 'store_true',
                # The default value must be None so CommandBase.get_options()
                # doesn't always overwrite the value with the config file value.
                'default': None,
            },
            ('--personal-rip', '--pr'): {
                'help': 'Tag as your own encode',
                'action': 'store_true',
            },
            ('--screenshots', '--ss'): {
                'help': ('How many screenshots to make '
                         f'(min={defaults["screenshots"].min}, '
                         f'max={defaults["screenshots"].max})'),
                'type': argtypes.number_of_screenshots(
                    min=defaults['screenshots'].min,
                    max=defaults['screenshots'].max,
                ),
            },
            ('--special', '--sp'): {
                'help': 'Tag as special episode, e.g. Christmas special (ignored for movie uploads)',
                'action': 'store_true',
            },
            ('--only-description', '--od'): {
                'help': 'Only generate description (do not upload anything)',
                'action': 'store_true',
                'group': 'generate-metadata',
            },
            ('--only-title', '--oti'): {
                'help': 'Only generate title (do not upload anything)',
                'action': 'store_true',
                'group': 'generate-metadata',
            },
        },
    }



"""
Concrete :class:`~.base.TrackerJobsBase` subclass for BHD
"""

import io
import os
import re

from ... import __homepage__, __project_name__, jobs
from ...utils import as_groups, cached_property, release, string
from ..base import TrackerJobsBase

import logging  # isort:skip
_log = logging.getLogger(__name__)


class BhdTrackerJobs(TrackerJobsBase):
    release_name_translation = {
        'audio_format': {
            # Don't match against end of string to preserve extensions like
            # "Atmos"
            re.compile(r'^AC-3'): r'DD',
            re.compile(r'^E-AC-3'): r'DDP',
        },
    }

    @cached_property
    def jobs_before_upload(self):
        # Make base class jobs conditional. Custom BHD jobs get their condition
        # in their properties.
        generic_job_attributes = ('imdb_job', 'tmdb_job', 'release_name_job',
                                  'scene_check_job', 'create_torrent_job', 'mediainfo_job',
                                  'screenshots_job', 'upload_screenshots_job',
                                  'add_torrent_job', 'copy_torrent_job')
        for job_attr in generic_job_attributes:
            job = getattr(self, job_attr, None)
            if job:
                job.condition = self.make_job_condition(job_attr)

        return (
            # Background jobs
            self.create_torrent_job,
            self.mediainfo_job,
            self.screenshots_job,
            self.upload_screenshots_job,

            # Interactive jobs
            self.imdb_job,
            self.tmdb_job,
            self.release_name_job,
            self.category_job,
            self.type_job,
            self.source_job,
            self.description_job,
            self.scene_check_job,
            self.tags_job,
        )

    def make_job_condition(self, job_attr):
        """
        Return :attr:`~.base.JobBase.condition` for jobs

        :param str job_attr: Name of the job attribute this condition is for
        """
        def condition():
            isolated_jobs = self.isolated_jobs
            if isolated_jobs:
                return job_attr in isolated_jobs
            else:
                return True

        return condition

    @property
    def isolated_jobs(self):
        """
        Sequence of attribute names (e.g. "imdb_job") that were singled out by the
        user, e.g. with a CLI argument
        """
        if self.options.get('only_description', False):
            # Only activate description_job and its dependencies
            return ('description_job', 'screenshots_job', 'upload_screenshots_job')
        elif self.options.get('only_title', False):
            # Only activate release_name_job and its dependencies
            return ('release_name_job', 'imdb_job')
        else:
            return ()

    @cached_property
    def imdb_job(self):
        imdb_job = super().imdb_job
        imdb_job.signal.register('output', self.handle_imdb_job_output)
        return imdb_job

    def handle_imdb_job_output(self, _):
        user_confirmed_type = self.imdb_job.query.type
        self.tmdb_job.query.type = user_confirmed_type

    @cached_property
    def category_job(self):
        # 'output' signal is only emitted when job succeeds while 'finished'
        # signal is also emitted when job fails (e.g. when Ctrl-c is pressed)
        self.release_name_job.signal.register('output', self.autodetect_category)
        return self.make_choice_job(
            name=self.get_job_name('category'),
            label='Category',
            condition=self.make_job_condition('category_job'),
            options=(
                {'label': 'Movie', 'value': '1'},
                {'label': 'TV', 'value': '2'},
            ),
        )

    _autodetect_category_map = {
        'Movie': lambda release_name: release_name.type is release.ReleaseType.movie,
        'TV': lambda release_name: release_name.type in (release.ReleaseType.season,
                                                         release.ReleaseType.episode)
    }

    def autodetect_category(self, _):
        approved_release_name = self.release_name
        _log.debug('Autodetecting category: Approved release type: %r', approved_release_name.type)
        for label, is_match in self._autodetect_category_map.items():
            if is_match(approved_release_name):
                # Focus autodetected choice
                self.category_job.focused = label
                # Get value of autodetected choice
                value = self.category_job.focused[1]
                # Mark autodetected choice
                self.category_job.set_label(value, f'{label} (autodetected)')
                break

    @cached_property
    def type_job(self):
        # 'output' signal is only emitted when job succeeds while 'finished'
        # signal is also emitted when job fails (e.g. when Ctrl-c is pressed)
        self.release_name_job.signal.register('output', self.autodetect_type)
        return self.make_choice_job(
            name=self.get_job_name('type'),
            label='Type',
            condition=self.make_job_condition('type_job'),
            options=(
                {'label': 'UHD 100', 'value': 'UHD 100'},
                {'label': 'UHD 66', 'value': 'UHD 66'},
                {'label': 'UHD 50', 'value': 'UHD 50'},
                {'label': 'UHD Remux', 'value': 'UHD Remux'},
                {'label': 'BD 50', 'value': 'BD 50'},
                {'label': 'BD 25', 'value': 'BD 25'},
                {'label': 'BD Remux', 'value': 'BD Remux'},
                {'label': '2160p', 'value': '2160p'},
                {'label': '1080p', 'value': '1080p'},
                {'label': '1080i', 'value': '1080i'},
                {'label': '720p', 'value': '720p'},
                {'label': '576p', 'value': '576p'},
                {'label': '540p', 'value': '540p'},
                {'label': 'DVD 9', 'value': 'DVD 9'},
                {'label': 'DVD 5', 'value': 'DVD 5'},
                {'label': 'DVD Remux', 'value': 'DVD Remux'},
                {'label': '480p', 'value': '480p'},
                {'label': 'Other', 'value': 'Other'},
            ),
        )

    _autodetect_type_map = {
        'DVD 9': lambda release_name: release_name.source == 'DVD9',
        'DVD 5': lambda release_name: release_name.source == 'DVD5',
        'DVD Remux': lambda release_name: release_name.source == 'DVD Remux',
        '2160p': lambda release_name: release_name.resolution == '2160p',
        '1080p': lambda release_name: release_name.resolution == '1080p',
        '1080i': lambda release_name: release_name.resolution == '1080i',
        '720p': lambda release_name: release_name.resolution == '720p',
        '576p': lambda release_name: release_name.resolution == '576p',
        '540p': lambda release_name: release_name.resolution == '540p',
        '480p': lambda release_name: release_name.resolution == '480p',
    }

    def autodetect_type(self, _):
        approved_release_name = self.release_name
        _log.debug('Autodetecting type: Approved resolution: %r', approved_release_name.resolution)
        _log.debug('Autodetecting type: Approved source: %r', approved_release_name.source)
        for label, is_match in self._autodetect_type_map.items():
            if is_match(approved_release_name):
                # Focus autodetected choice
                self.type_job.focused = label
                # Get value of autodetected choice
                value = self.type_job.focused[1]
                # Mark autodetected choice
                self.type_job.set_label(value, f'{label} (autodetected)')
                # Select autodetected choice (i.e. finish job and don't prompt user)
                self.type_job.choice = value
                break
        else:
            self.type_job.focused = 'Other'

    @cached_property
    def source_job(self):
        # 'output' signal is only emitted when job succeeds while 'finished'
        # signal is also emitted when job fails (e.g. when Ctrl-c is pressed)
        self.release_name_job.signal.register('output', self.autodetect_source)
        return self.make_choice_job(
            name=self.get_job_name('source'),
            label='Source',
            condition=self.make_job_condition('source_job'),
            options=(
                {'label': 'Blu-ray', 'value': 'Blu-ray'},
                {'label': 'HD-DVD', 'value': 'HD-DVD'},
                {'label': 'WEB', 'value': 'WEB'},
                {'label': 'HDTV', 'value': 'HDTV'},
                {'label': 'DVD', 'value': 'DVD'},
            ),
        )

    # Map type_job labels to matchers
    _autodetect_source_map = {
        'Blu-ray': lambda release_name: 'BluRay' in release_name.source,
        'HD-DVD': lambda release_name: 'HD-DVD' in release_name.source,
        'WEB': lambda release_name: 'WEB' in release_name.source,
        'HDTV': lambda release_name: 'HDTV' in release_name.source,
        'DVD': lambda release_name: 'DVD' in release_name.source,
    }

    def autodetect_source(self, _):
        approved_release_name = self.release_name
        _log.debug('Autodetecting source: Approved source: %r', approved_release_name.source)
        for label, is_match in self._autodetect_source_map.items():
            if is_match(approved_release_name):
                # Focus autodetected choice
                self.source_job.focused = label
                # Get value of autodetected choice
                value = self.source_job.focused[1]
                # Mark autodetected choice
                self.source_job.set_label(value, f'{label} (autodetected)')
                # Select autodetected choice (i.e. finish job and don't prompt user)
                self.source_job.choice = value
                break

    @cached_property
    def description_job(self):
        job = jobs.dialog.TextFieldJob(
            name=self.get_job_name('description'),
            label='Description',
            condition=self.make_job_condition('description_job'),
            read_only=True,
            **self.common_job_args,
        )
        job.add_task(
            job.fetch_text(
                coro=self.generate_description(),
                finish_on_success=True,
            )
        )
        return job

    image_host_config = {
        'common': {'thumb_width': 350},
    }

    async def generate_description(self):
        # Wait until all screenshots are uploaded
        await self.upload_screenshots_job.wait()
        rows = []
        screenshot_groups = as_groups(
            self.upload_screenshots_job.uploaded_images,
            group_sizes=(2,),
            default=None,
        )
        for screenshots in screenshot_groups:
            cells = []
            for screenshot in screenshots:
                if screenshot is not None:
                    if screenshot.thumbnail_url is None:
                        raise RuntimeError(f'No thumbnail for {screenshot}')
                    cells.append(f'[url={screenshot}][img]{screenshot.thumbnail_url}[/img][/url]')
            # Space between columns
            rows.append(' '.join(cells))
        screenshots = '[center]\n' + '\n\n'.join(rows) + '\n[/center]'
        promotion = (
            '[right][size=1]'
            f'Shared with [url={__homepage__}]{__project_name__}[/url]'
            '[/size][/right]'
        )
        return screenshots + '\n\n' + promotion

    @cached_property
    def tags_job(self):
        job = jobs.dialog.TextFieldJob(
            name=self.get_job_name('tags'),
            label='Tags',
            condition=self.make_job_condition('tags_job'),
            read_only=True,
            **self.common_job_args,
        )
        job.add_task(
            job.fetch_text(
                coro=self.autodetect_tags(),
                finish_on_success=True,
            )
        )
        return job

    async def autodetect_tags(self):
        await self.release_name_job.wait()
        await self.scene_check_job.wait()

        # Any additional tags separated by comma(s). (Commentary, 2in1, Hybrid,
        # OpenMatte, 2D3D, WEBRip, WEBDL, 3D, 4kRemaster, DualAudio, EnglishDub,
        # Personal, Scene, DigitalExtras, Extras)
        tags = []
        if 'WEBRip' in self.release_name.source:
            tags.append('WEBRip')
        elif 'WEB-DL' in self.release_name.source:
            tags.append('WEBDL')
        if 'Hybrid' in self.release_name.source:
            tags.append('Hybrid')
        if self.release_name.has_commentary:
            tags.append('Commentary')
        if self.release_name.has_dual_audio:
            tags.append('DualAudio')
        if 'Open Matte' in self.release_name.edition:
            tags.append('OpenMatte')
        if '2in1' in self.release_name.edition:
            tags.append('2in1')
        if '4k Remastered' in self.release_name.edition:
            tags.append('4kRemaster')
        if self.get_job_attribute(self.scene_check_job, 'is_scene_release'):
            tags.append('Scene')
        if self.options['personal_rip']:
            tags.append('Personal')

        # TODO: 2D3D
        # TODO: 3D
        # TODO: EnglishDub
        # TODO: DigitalExtras
        # TODO: Extras

        return '\n'.join(tags)

    @property
    def submission_ok(self):
        """
        `False` if :attr:`~.TrackerJobsBase.options` prevents submission for any
        reason, parent class implementation otherwise
        """
        if self.isolated_jobs:
            return False
        else:
            return super().submission_ok

    @property
    def post_data(self):
        return {
            'name': self.get_job_output(self.release_name_job, slice=0),
            'category_id': self.get_job_attribute(self.category_job, 'choice'),
            'type': self.get_job_attribute(self.type_job, 'choice'),
            'source': self.get_job_attribute(self.source_job, 'choice'),
            'imdb_id': self.get_job_output(self.imdb_job, slice=0),
            'tmdb_id': self.get_job_output(self.tmdb_job, slice=0).split('/')[1],
            'description': self.get_job_output(self.description_job, slice=0),
            'edition': self.post_data_edition,
            'custom_edition': self.options['custom_edition'],
            'tags': ','.join(self.get_job_output(self.tags_job, slice=0).split('\n')),
            'nfo': self.post_data_nfo,
            'pack': self.post_data_pack,
            'sd': self.post_data_sd,
            'special': self.post_data_special,
            'anon': '1' if self.options['anonymous'] else '0',
            'live': '0' if self.options['draft'] else '1',
        }

    @cached_property
    def post_data_edition(self):
        # The edition of the uploaded release. (Collector, Director, Extended,
        # Limited, Special, Theatrical, Uncut or Unrated)
        edition = self.release_name.edition
        _log.debug('Approved edition: %r', edition)
        if "Collector's Edition" in edition:
            return 'Collector'
        elif "Director's Cut" in edition:
            return 'Director'
        elif 'Extended Cut' in edition:
            return 'Extended'
        elif 'Limited' in edition:
            return 'Limited'
        elif 'Special Edition' in edition:
            return 'Special'
        elif 'Theatrical Cut' in edition:
            return 'Theatrical'
        elif 'Uncut' in edition or 'Uncensored' in edition:
            return 'Uncut'
        elif 'Unrated' in edition:
            return 'Unrated'

    @property
    def post_data_pack(self):
        # The TV pack flag for when the torrent contains a complete season.
        # (0 = No TV pack or 1 = TV Pack). Default is 0
        if self.release_name.type is release.ReleaseType.season:
            return '1'
        else:
            return '0'

    @property
    def post_data_sd(self):
        # The SD flag. (0 = Not Standard Definition, 1 = Standard Definition).
        # Default is 0
        try:
            height = int(self.release_name.resolution[:-1])
        except ValueError:
            return '0'
        else:
            return '1' if height < 720 else '0'

    max_nfo_size = 500_000

    @property
    def post_data_nfo(self):
        # The NFO of the torrent as string.
        if os.path.isdir(self.content_path):
            for entry in os.listdir(self.content_path):
                if entry.lower().endswith('.nfo'):
                    nfo_path = os.path.join(self.content_path, entry)
                    # Limit size to 500kB
                    if os.path.getsize(nfo_path) <= self.max_nfo_size:
                        try:
                            with open(nfo_path, 'rb') as f:
                                return string.autodecode(f.read())
                        except OSError as e:
                            self.error(e.strerror if e.strerror else str(e))

    @property
    def post_data_special(self):
        # The TV special flag for when the torrent contains a TV special. (0 =
        # Not a TV special, 1 = TV Special). Default is 0
        if self.release_name.type is release.ReleaseType.episode:
            if self.options['special']:
                return '1'
        return '0'

    # TODO
    # @property
    # def post_data_region(self):
    #     # The region in which the disc was released. Only for discs! (AUS,
    #     # CAN, CEE, CHN, ESP, EUR, FRA, GBR, GER, HKG, ITA, JPN, KOR, NOR,
    #     # NLD, RUS, TWN or USA)

    @property
    def torrent_filepath(self):
        return self.get_job_output(self.create_torrent_job, slice=0)

    @property
    def mediainfo_filehandle(self):
        mediainfo = self.get_job_output(self.mediainfo_job, slice=0)
        return io.BytesIO(bytes(mediainfo, 'utf-8'))



from ....utils import cached_property, fs
from .. import utils, widgets
from . import JobWidgetBase

import logging  # isort:skip
_log = logging.getLogger(__name__)


class ScreenshotsJobWidget(JobWidgetBase):
    def setup(self):
        content_name = fs.basename(self.job.kwargs['content_path'])
        self._throbber = utils.Throbber(
            format=f'Analyzing {content_name} {{throbber}}',
            callback=self.handle_throbber_state,
            active=True,
        )
        self._screenshot_progress = widgets.ProgressBar()
        self.job.signal.register('output', self.handle_screenshot_path)
        self.job.signal.register('finished', lambda _: self.invalidate())

    def handle_throbber_state(self, state):
        self.job.info = state

    def handle_screenshot_path(self, path):
        self._throbber.active = False
        self.job.info = ''
        if self.job.screenshots_total > 0:
            self._screenshot_progress.percent = self.job.screenshots_created / self.job.screenshots_total * 100

    @cached_property
    def runtime_widget(self):
        return self._screenshot_progress



"""
Classes and functions that are used by all :class:`~.base.WebDbApiBase`
subclasses
"""

import re

from .. import release, signal
from ..types import ReleaseType

import logging  # isort:skip
_log = logging.getLogger(__name__)


class Query:
    """
    Search query for databases like IMDb

    :param str title: Name of the movie or TV series
    :param type: :class:`~.types.ReleaseType` enum or one of its value names
    :param year: Year of release
    :type year: str or int
    :param str id: Known ID for a specific DB

    :raise ValueError: if an invalid argument is passed
    """

    @staticmethod
    def _normalize_title(title):
        return ' '.join(title.casefold().strip().split())

    _kwarg_defaults = {
        'year': None,
        'type': ReleaseType.unknown,
        'id': None,
    }

    def __init__(self, title='', **kwargs):
        self._kwargs_order = tuple(kwargs)
        self._signal = signal.Signal()
        self._signal.add('changed')
        self.title = title
        self.type = kwargs.get('type', self._kwarg_defaults['type'])
        self.year = kwargs.get('year', self._kwarg_defaults['year'])
        self.id = kwargs.get('id', self._kwarg_defaults['id'])

    @property
    def type(self):
        """:class:`~.types.ReleaseType` value"""
        return self._type

    @type.setter
    def type(self, type):
        before = getattr(self, '_type', None)

        self._type = ReleaseType(type)

        if self._type != before:
            self.signal.emit('changed', self)

    @property
    def title(self):
        """Name of the movie or TV series"""
        return self._title

    @title.setter
    def title(self, title):
        before = getattr(self, '_title_normalized', None)

        self._title = str(title)
        self._title_normalized = self._normalize_title(self.title)

        if self._title_normalized != before:
            self.signal.emit('changed', self)

    @property
    def title_normalized(self):
        """Same as :attr:`title` but in stripped lower case with deduplicated spaces"""
        return self._title_normalized

    @property
    def year(self):
        """Year of release"""
        return self._year

    @year.setter
    def year(self, year):
        before = getattr(self, '_year', None)

        if year is None:
            self._year = None
        else:
            try:
                year_int = int(year)
            except (TypeError, ValueError):
                raise ValueError(f'Invalid year: {year}')
            else:
                if not 1800 < year_int < 2100:
                    raise ValueError(f'Invalid year: {year}')
                else:
                    self._year = str(year_int)

        if self._year != before:
            self.signal.emit('changed', self)

    @property
    def id(self):
        """Known ID for a specific DB"""
        return self._id

    @id.setter
    def id(self, id):
        before = getattr(self, '_id', None)

        self._id = None if id is None else str(id)

        if self._id != before:
            self.signal.emit('changed', self)

    def update(self, query):
        """
        Copy property values from other query

        :param query: :class:`Query` instance to copy values from
        """
        before = str(self)

        with self.signal.suspend('changed'):
            for attr in ('title', 'type', 'year', 'id'):
                other_value = getattr(query, attr)
                setattr(self, attr, other_value)

        if str(self) != before:
            self.signal.emit('changed', self)

    _types = {
        ReleaseType.movie: ('movie', 'film'),
        ReleaseType.season: ('season', 'series', 'tv', 'show', 'tvshow'),
        ReleaseType.episode: ('episode',),
    }
    _known_types = tuple(v for types in _types.values() for v in types)
    _kw_regex = {
        'year': r'year:(\d{4})',
        'type': rf'type:({"|".join(_known_types)})',
        'id': r'id:(\S+)',
    }

    @classmethod
    def from_string(cls, query):
        """
        Create instance from string

        The returned :class:`Query` is case-insensitive and has any superfluous
        whitespace removed.

        Keyword arguments are extracted by looking for ``"year:YEAR"``,
        ``"type:TYPE"`` and ``"id:ID"`` in `query` where ``YEAR`` is a
        four-digit number, ``TYPE`` is something like "movie", "film", "tv", etc
        and ``ID`` is a known ID for the DB this query is meant for.
        """
        def get_kwarg(string):
            for kw, regex in cls._kw_regex.items():
                match = re.search(f'^{regex}$', part)
                if match:
                    value = match.group(1)
                    if kw == 'type' and value in cls._types[ReleaseType.movie]:
                        return 'type', ReleaseType.movie
                    elif kw == 'type' and value in cls._types[ReleaseType.season]:
                        return 'type', ReleaseType.season
                    elif kw == 'type' and value in cls._types[ReleaseType.episode]:
                        return 'type', ReleaseType.episode
                    elif kw == 'year':
                        return 'year', value
                    elif kw == 'id':
                        return 'id', value
            return None, None

        # Extract "key:value" pairs (e.g. "year:2015")
        title = []
        kwargs = {}
        for part in str(query).strip().split():
            kw, value = get_kwarg(part)
            if (kw, value) != (None, None):
                kwargs[kw] = value
            else:
                title.append(part)
        if title:
            kwargs['title'] = ' '.join(title)
        return cls(**kwargs)

    @classmethod
    def from_path(cls, path):
        """
        Create instance from file or directory name

        `path` is passed to :class:`~.release.ReleaseInfo` to get the
        arguments for instantiation.
        """
        info = release.ReleaseInfo(path)
        kwargs = {'title': info['title']}
        if info.get('year'):
            kwargs['year'] = info['year']
        if info.get('type'):
            kwargs['type'] = info['type']
        return cls(**kwargs)

    @property
    def signal(self):
        """
        :class:`~.signal.Signal` instance

        Available signals:

        ``changed``
            Emitted after query parameters changed. Registered callbacks get the
            instance as a positional argument.
        """
        return self._signal

    def __eq__(self, other):
        if isinstance(other, type(self)):
            return (
                self.title_normalized == other.title_normalized
                and self.year == other.year
                and self.type is other.type
                and self.id == other.id
            )
        else:
            return NotImplemented

    def __str__(self):
        if self.id:
            return f'id:{self.id}'
        else:
            parts = [self.title]
            for attr in self._kwargs_order:
                value = getattr(self, attr)
                if value:
                    parts.append(f'{attr}:{value}')
            return ' '.join(parts)

    def __repr__(self):
        kwargs = ', '.join(
            f'{k}={v!r}'
            for k, v in (('title', self.title),
                         ('year', self.year),
                         ('type', self.type),
                         ('id', self.id))
            if v
        )
        return f'{type(self).__name__}({kwargs})'


class SearchResult:
    """
    Information about a search result

    Keyword arguments are available as attributes.

    `cast`, `countries`, `directors`, `genres`, `summary`, `title_english` and
    `title_original` may be coroutine functions that return the expected value.

    :param str directors: Sequence of directors
    :param cast: Short list of actor names
    :type cast: sequence of str
    :param str countries: List of country names of origin
    :param str id: ID for the relevant DB
    :param genres: Short list of genres, e.g. `["horror", "commedy"]`
    :type genres: sequence of str
    :param summary: Short text that describes the movie or series
    :param str title: Title of the movie or series
    :param str title_english: English title of the movie or series
    :param str title_original: Original title of the movie or series
    :param str type: :class:`~.types.ReleaseType` value
    :param str url: Web page of the search result
    :param str year: Release year; for series this should be the year of the
        first airing of the first episode of the first season
    """
    def __init__(self, *, id, type, url, year, cast=(), countries=(), directors='',
                 genres=(), summary='', title, title_english='',
                 title_original=''):
        self._info = {
            'id' : id,
            'title' : str(title),
            'type' : ReleaseType(type),
            'url' : str(url),
            'year' : str(year),
            # These may be coroutine functions
            'cast' : cast,
            'countries' : countries,
            'directors' : directors,
            'genres' : genres,
            'summary' : summary,
            'title_english' : title_english,
            'title_original' : title_original,
        }

    def __getattr__(self, name):
        try:
            return self._info[name]
        except KeyError:
            raise AttributeError(name)

    def __repr__(self):
        kwargs = ', '.join(f'{k}={v!r}' for k, v in self._info.items())
        return f'{type(self).__name__}({kwargs})'


class Person(str):
    """:class:`str` subclass with an `url` attribute"""

    def __new__(cls, name, url=''):
        obj = super().__new__(cls, name)
        obj.url = url
        return obj



"""
Classes and functions that are not specific to
:class:`~.base.SceneDbApiBase` subclasses
"""

from ..types import ReleaseType

_needed_movie_keys = ('title', 'year', 'resolution', 'source', 'video_codec', 'group')
_needed_series_keys = ('title', 'episodes', 'resolution', 'source', 'video_codec', 'group')


def get_needed_keys(release_info):
    """
    Return needed :class:`~.release.ReleaseInfo` keys to identify release

    :param release_info: :class:`~.release.ReleaseInfo` instance or any
        :class:`dict`-like object with the keys ``type`` and ``source``

    :return: Sequence of required keys or empty sequence if `release_info`
        doesn't contain a ``type``
    """
    if release_info['type'] is ReleaseType.movie:
        needed_keys = _needed_movie_keys
    elif release_info['type'] in (ReleaseType.season, ReleaseType.episode):
        needed_keys = _needed_series_keys
    else:
        # If we don't even know the type, we certainly don't have enough
        # information to pin down a release.
        return ()

    # DVDRips typically don't include resolution in release name
    if release_info['source'] == 'DVDRip':
        needed_keys = list(needed_keys)
        needed_keys.remove('resolution')

    return tuple(needed_keys)



from prompt_toolkit.filters import Condition
from prompt_toolkit.layout.containers import (ConditionalContainer,
                                              HorizontalAlign, HSplit, VSplit,
                                              Window, WindowAlign)
from prompt_toolkit.layout.controls import FormattedTextControl
from prompt_toolkit.layout.dimension import Dimension

from ....utils import cached_property, fs
from .. import utils, widgets
from . import JobWidgetBase

import logging  # isort:skip
_log = logging.getLogger(__name__)


class CreateTorrentJobWidget(JobWidgetBase):
    def setup(self):
        self._progress = widgets.ProgressBar()
        self._file_tree = FormattedTextControl()
        self._bytes_per_second = FormattedTextControl()
        self._percent_done = FormattedTextControl()
        self._seconds_elapsed = FormattedTextControl()
        self._seconds_remaining = FormattedTextControl()
        self._seconds_total = FormattedTextControl()
        self._time_started = FormattedTextControl()
        self._time_finished = FormattedTextControl()

        status_column1 = HSplit(
            children=[
                self.make_readout(' Started:', '_time_started', align='left', width=8),
                self.make_readout('Finished:', '_time_finished', align='left', width=8),
            ],
        )
        status_column2 = HSplit(
            children=[
                self.make_readout('  Elapsed:', '_seconds_elapsed', align='center', width=8),
                self.make_readout('Remaining:', '_seconds_remaining', align='center', width=8),
                self.make_readout('    Total:', '_seconds_total', align='center', width=8),
            ],
        )
        status_column3 = HSplit(
            children=[
                self.make_readout('%'.ljust(5), '_percent_done', align='right', width=6, label_side='right'),
                self.make_readout(
                    lambda: (self._bytes_per_second_string.split(' ')[1] + '/s').ljust(5),
                    '_bytes_per_second',
                    align='right',
                    width=6,
                    label_side='right',
                ),
            ],
        )
        status = VSplit(children=[status_column1, status_column2, status_column3], padding=3)

        self._runtime_widget = HSplit(
            children=[
                self._progress,
                ConditionalContainer(
                    filter=Condition(lambda: self._progress.percent > 0),
                    content=HSplit([status, Window(self._file_tree)]),
                ),
            ],
            style='class:info',
            width=Dimension(min=55, max=100),
        )

        self._throbber = utils.Throbber(
            callback=self.handle_throbber_state,
            format='Getting announce URL {throbber}'
        )
        self.job.signal.register('announce_url', self.handle_announce_url)
        self.job.signal.register('file_tree', self.handle_file_tree)
        self.job.signal.register('progress_update', self.handle_progress_update)
        self.job.signal.register('finished', lambda _: self.invalidate())

    def make_readout(self, label, attribute, align, width, label_side='left'):
        readout_widget = Window(
            getattr(self, attribute),
            dont_extend_width=True,
            dont_extend_height=True,
            style='class:info.readout',
            width=width,
            align=WindowAlign.RIGHT,
        )
        label_widget = Window(FormattedTextControl(label), dont_extend_width=True)

        if label_side == 'left':
            children = [label_widget, readout_widget]
        elif label_side == 'right':
            children = [readout_widget, label_widget]

        return VSplit(
            children=children,
            padding=1,
            align=getattr(HorizontalAlign, align.upper()),
        )

    def handle_throbber_state(self, state):
        self.job.info = state
        self.invalidate()

    def handle_file_tree(self, file_tree):
        self._file_tree.text = fs.format_file_tree(file_tree)

    def handle_announce_url(self, announce_url):
        if announce_url is Ellipsis:
            self._throbber.active = True
        else:
            self._throbber.active = False

    def handle_progress_update(self, status):
        self._progress.percent = status.percent_done

        self._bytes_per_second_string = status.bytes_per_second.format(
            prefix='binary',
            decimal_places=2,
            trailing_zeros=True,
        )
        self._bytes_per_second.text = self._bytes_per_second_string.split(' ')[0]

        self._percent_done.text = f'{status.percent_done:.2f}'

        self._time_started.text = status.time_started.strftime('%H:%M:%S')
        self._time_finished.text = status.time_finished.strftime('%H:%M:%S')

        def timedelta_string(delta):
            hours, rest = divmod(delta.total_seconds(), 3600)
            minutes, seconds = divmod(rest, 60)
            return f'{hours:02.0f}:{minutes:02.0f}:{seconds:02.0f}'

        self._seconds_elapsed.text = timedelta_string(status.seconds_elapsed)
        self._seconds_remaining.text = timedelta_string(status.seconds_remaining)
        self._seconds_total.text = timedelta_string(status.seconds_total)

        self.invalidate()

    @cached_property
    def runtime_widget(self):
        return self._runtime_widget


class AddTorrentJobWidget(JobWidgetBase):
    def setup(self):
        pass

    @cached_property
    def runtime_widget(self):
        return None


class CopyTorrentJobWidget(JobWidgetBase):
    def setup(self):
        pass

    @cached_property
    def runtime_widget(self):
        return None



"""
Wrapper for ``mediainfo`` command
"""

from .. import errors
from ..utils import fs, get_aioloop, video
from . import JobBase

import logging  # isort:skip
_log = logging.getLogger(__name__)


class MediainfoJob(JobBase):
    """
    Get output from ``mediainfo`` command

    See :func:`.utils.video.mediainfo` for more information.
    """

    name = 'mediainfo'
    label = 'Mediainfo'

    # Don't show mediainfo output in TUI. It is printed out to stdout if this is
    # the only/final job.
    hidden = True

    @property
    def cache_id(self):
        """Final segment of `content_path`"""
        return fs.basename(self._content_path)

    def initialize(self, *, content_path):
        """
        Set internal state

        :param content_path: Path to video file or directory that contains a
            video file
        """
        self._content_path = content_path

    def execute(self):
        """Call ``mediainfo`` in an asynchronous thread"""
        self.add_task(
            coro=self._get_mediainfo(),
            finish_when_done=True,
        )

    async def _get_mediainfo(self):
        loop = get_aioloop()
        try:
            mediainfo = await loop.run_in_executor(None, lambda: video.mediainfo(self._content_path))
        except errors.ContentError as e:
            self.error(e)
        else:
            self.send(mediainfo)



"""
Find files from a torrent and create links if necessary
"""

from .... import jobs, utils
from .base import CommandBase


class download_location(CommandBase):
    """
    Hard-link existing files as TORRENT expects them and print their location

    Recursively find all files beneath each LOCATION that have the same size as
    those specified in TORRENT and hash two pieces of each size-matching file to
    be extra sure.

    For each file in TORRENT, hard-link the first match to the file path
    specified in TORRENT, prepending LOCATION to the relative path from the
    torrent.

    Print the first LOCATION beneath which a file from TORRENT is found. This is
    the download path of the TORRENT you should pass to your client when adding
    the torrent.

    If no matching files are found, print the location passed to --default or
    the first LOCATION if there is no --default location.
    """

    names = ('download-location', 'dloc')

    argument_definitions = {
        'TORRENT': {
            'help': 'Path to torrent file',
        },
        'LOCATION': {
            'nargs': '+',
            'help': 'Potential path of existing files in TORRENT',
        },
        ('--default',): {
            'help': 'Default location if no existing files are found',
        },
    }

    @utils.cached_property
    def jobs(self):
        return (
            jobs.download_location.DownloadLocationJob(
                home_directory=self.home_directory,
                cache_directory=self.cache_directory,
                torrent=self.args.TORRENT,
                locations=self.args.LOCATION,
                default=self.args.default,
            ),
        )



"""
Generate uniform release name
"""

from .. import errors
from ..utils import fs, release
from . import JobBase

import logging  # isort:skip
_log = logging.getLogger(__name__)


class ReleaseNameJob(JobBase):
    """
    Create user-confirmed release name from file or directory path

    This job adds the following signals to the :attr:`~.JobBase.signal`
    attribute:

        ``release_name_updating``
            Emitted when :meth:`~.ReleaseName.fetch_info` is called. Registered
            callbacks get no arguments.

        ``release_name_updated``
            Emitted after :meth:`~.ReleaseName.fetch_info` updated the release
            name. Registered callbacks get a :class:`~.release.ReleaseName`
            instance as a positional argument.

        ``release_name``
            Emitted after the user approved a release name. Registered callbacks
            get a :class:`~.release.ReleaseName` instance as a positional
            argument.
    """

    name = 'release-name'
    label = 'Release Name'

    @property
    def cache_id(self):
        """Final segment of `content_path`"""
        return fs.basename(self._content_path)

    @property
    def release_name(self):
        """:class:`~.utils.release.ReleaseName` instance"""
        return self._release_name

    def initialize(self, *, content_path):
        """
        Set internal state

        :param content_path: Path to release file or directory
        """
        self._content_path = content_path
        self._release_name = release.ReleaseName(content_path)
        self.signal.add('release_name_updating')
        self.signal.add('release_name_updated')
        self.signal.add('release_name')

    def release_name_selected(self, release_name):
        """
        :meth:`Send <upsies.jobs.base.JobBase.send>` release name and :meth:`finish
        <upsies.jobs.base.JobBase.finish>`

        This method must be called by the UI when the user accepts the release
        name.

        :param str release_name: Release name as accepted by the user
        """
        # Parse string to get ReleaseName instance for the "release_name" signal
        self._release_name = release.ReleaseName(release_name)
        self.signal.emit('release_name', self.release_name)
        # Send original string as we received it
        self.send(release_name)
        self.finish()

    def fetch_info(self, *args, **kwargs):
        """
        Synchronous wrapper around :meth:`.ReleaseName.fetch_info` that emits
        ``release_name_updating`` before the info is fetched and
        ``release_name_updated`` afterwards.

        All arguments are passed on to :meth:`.ReleaseName.fetch_info`.
        """
        self.signal.emit('release_name_updating')

        def fetch_info_done(_):
            self.signal.emit('release_name_updated', self.release_name)

        self.add_task(
            self._catch_errors(
                self.release_name.fetch_info(*args, **kwargs),
            ),
            callback=fetch_info_done,
        )

    async def _catch_errors(self, coro):
        try:
            return await coro
        except errors.RequestError as e:
            self.error(e)



"""
TUI representation of :class:`jobs <upsies.jobs.base.JobBase>`
"""

from ....utils import subclasses, submodules
from .base import JobWidgetBase


def JobWidget(job, app):
    """
    Factory that returns JobWidgetBase instances based on job type

    The widget class name is created by adding "Widget" to `job`'s class name.
    The widget class is imported from :mod:`.jobwidgets`.

    :param job: Job instance
    :type job: :class:`~.jobs.base.JobBase`
    :param app: Application instance
    :type job: :class:`~.prompt_toolkit.application.Application`

    :raise RuntimeError: if `job`'s type is not supported
    """
    widget_cls_name = type(job).__name__ + 'Widget'
    widget_clses = subclasses(JobWidgetBase, submodules(__package__))
    for widget_cls in widget_clses:
        if widget_cls.__name__ == widget_cls_name:
            return widget_cls(job, app)
    raise RuntimeError(f'No widget class found for job: {job!r}')



"""
Upload images to image hosting service
"""

from .... import jobs, utils
from .base import CommandBase


class upload_images(CommandBase):
    """Upload images to image hosting service"""

    names = ('upload-images', 'ui')

    argument_definitions = {}

    subcommand_name = 'IMGHOST'
    subcommands = {
        imghost.name: {
            'description': imghost.description,
            'cli': {
                # Default arguments for all image hosts
                **{
                    'IMAGE': {
                        'nargs': '+',
                        'help': 'Path to image file',
                    },
                    ('--thumb-width', '-t'): {
                        'help': 'Thumbnail width in pixels',
                        'type': utils.argtypes.integer,
                        'default': None,
                    },
                },
                # Custom arguments defined by image host
                **imghost.argument_definitions,
            },
        }
        for imghost in utils.imghosts.imghosts()
    }

    @utils.cached_property
    def imghost_name(self):
        """Lower-case image host name"""
        return self.args.subcommand.lower()

    @utils.cached_property
    def imghost_options(self):
        """
        Relevant section in image host configuration file combined with CLI
        arguments where CLI arguments take precedence unless their value is
        `None`
        """
        return self.get_options('imghosts', self.imghost_name)

    @utils.cached_property
    def jobs(self):
        return (
            jobs.imghost.ImageHostJob(
                home_directory=self.home_directory,
                cache_directory=self.cache_directory,
                ignore_cache=self.args.ignore_cache,
                imghost=utils.imghosts.imghost(
                    name=self.imghost_name,
                    options=self.imghost_options,
                ),
                enqueue=self.args.IMAGE,
            ),
        )



from base64 import b64decode

from ... import errors
from .. import http
from . import base

import logging  # isort:skip
_log = logging.getLogger(__name__)


class PreDbApi(base.SceneDbApiBase):
    name = 'predb'
    label = 'PreDB'

    default_config = {}

    _url_base = b64decode('cHJlZGIub3Zo').decode('ascii')
    _search_url = f'https://{_url_base}/api/v1/'

    async def _search(self, keywords, group):
        if group:
            keywords = list(keywords)
            keywords.extend(('@team', str(group).replace('@', r'\@')))
        params = {'q': ' '.join((kw.lower() for kw in keywords)), 'count': 1000}
        _log.debug('Scene search: %r, %r', self._search_url, params)

        response = (await http.get(self._search_url, params=params, cache=True)).json()

        # Report API error or return list of release names
        if response['status'] != 'success':
            raise errors.RequestError(f'{self.label}: {response["message"]}')
        else:
            return (result['name'] for result in response['data']['rows'])

    async def release_files(self, release_name):
        """Always return an empty :class:`dict`"""
        return {}



"""
Connect the user interface to the engine room
"""

from .base import JobBase, QueueJobBase  # isort:skip
from . import (config, custom, dialog, download_location, imghost, mediainfo,
               release_name, scene, screenshots, submit, torrent, webdb)
